#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import os
import platform
import sys

import click
import urllib3
from bravado.exception import HTTPNotFound
from bravado.requests_client import RequestsClient
from packaging import version

from neptune.common.exceptions import STYLES
from neptune.common.oauth import NeptuneAuthenticator
from neptune.common.utils import (
    NoopObject,
    update_session_proxies,
)
from neptune.internal.backends.hosted_client import NeptuneResponseAdapter
from neptune.legacy.api_exceptions import (
    ProjectNotFound,
    WorkspaceNotFound,
)
from neptune.legacy.backend import (
    BackendApiClient,
    LeaderboardApiClient,
)
from neptune.legacy.exceptions import UnsupportedClientVersion
from neptune.legacy.internal.api_clients.credentials import Credentials
from neptune.legacy.internal.api_clients.hosted_api_clients.hosted_alpha_leaderboard_api_client import (
    HostedAlphaLeaderboardApiClient,
)
from neptune.legacy.internal.api_clients.hosted_api_clients.mixins import HostedNeptuneMixin
from neptune.legacy.internal.api_clients.hosted_api_clients.utils import legacy_with_api_exceptions_handler
from neptune.legacy.projects import Project

_logger = logging.getLogger(__name__)


class HostedNeptuneBackendApiClient(HostedNeptuneMixin, BackendApiClient):
    @legacy_with_api_exceptions_handler
    def __init__(self, api_token=None, proxies=None):
        self._old_leaderboard_client = None
        self._new_leaderboard_client = None
        self._api_token = api_token
        self._proxies = proxies

        # This is not a top-level import because of circular dependencies
        from neptune import __version__

        self.client_lib_version = __version__

        self.credentials = Credentials(api_token)

        ssl_verify = True
        if os.getenv("NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE"):
            urllib3.disable_warnings()
            ssl_verify = False

        self._http_client = RequestsClient(ssl_verify=ssl_verify, response_adapter_class=NeptuneResponseAdapter)
        # for session re-creation we need to keep an authenticator-free version of http client
        self._http_client_for_token = RequestsClient(
            ssl_verify=ssl_verify, response_adapter_class=NeptuneResponseAdapter
        )

        user_agent = "neptune-client/{lib_version} ({system}, python {python_version})".format(
            lib_version=self.client_lib_version,
            system=platform.platform(),
            python_version=platform.python_version(),
        )
        self.http_client.session.headers.update({"User-Agent": user_agent})
        self._http_client_for_token.session.headers.update({"User-Agent": user_agent})

        update_session_proxies(self.http_client.session, proxies)
        update_session_proxies(self._http_client_for_token.session, proxies)

        config_api_url = self.credentials.api_url_opt or self.credentials.token_origin_address
        # We don't need to be able to resolve Neptune host if we use proxy
        if proxies is None:
            self._verify_host_resolution(config_api_url, self.credentials.token_origin_address)

        # this backend client is used only for initial configuration and session re-creation
        self.backend_client = self._get_swagger_client(
            "{}/api/backend/swagger.json".format(config_api_url),
            self._http_client_for_token,
        )
        self._client_config = self._create_client_config(
            api_token=self.credentials.api_token, backend_client=self.backend_client
        )

        self._verify_version()

        self.backend_swagger_client = self._get_swagger_client(
            "{}/api/backend/swagger.json".format(self._client_config.api_url),
            self.http_client,
        )

        self.authenticator = self._create_authenticator(
            api_token=self.credentials.api_token,
            ssl_verify=ssl_verify,
            proxies=proxies,
            backend_client=self.backend_client,
        )
        self.http_client.authenticator = self.authenticator

        if sys.version_info >= (3, 7):
            try:
                os.register_at_fork(after_in_child=self._handle_fork_in_child)
            except AttributeError:
                pass

    def _handle_fork_in_child(self):
        self.backend_swagger_client = NoopObject()

    @property
    def api_address(self):
        return self._client_config.api_url

    @property
    def http_client(self):
        return self._http_client

    @property
    def display_address(self):
        return self._client_config.display_url

    @property
    def proxies(self):
        return self._proxies

    @legacy_with_api_exceptions_handler
    def get_project(self, project_qualified_name):
        try:
            response = self.backend_swagger_client.api.getProject(projectIdentifier=project_qualified_name).response()
            warning = response.metadata.headers.get("X-Server-Warning")
            if warning:
                click.echo("{warning}{content}{end}".format(content=warning, **STYLES))
            project = response.result

            return Project(
                backend=self.create_leaderboard_backend(project=project),
                internal_id=project.id,
                namespace=project.organizationName,
                name=project.name,
            )
        except HTTPNotFound:
            raise ProjectNotFound(project_qualified_name)

    @legacy_with_api_exceptions_handler
    def get_projects(self, namespace):
        try:
            r = self.backend_swagger_client.api.listProjects(organizationIdentifier=namespace).response()
            return r.result.entries
        except HTTPNotFound:
            raise WorkspaceNotFound(namespace_name=namespace)

    def create_leaderboard_backend(self, project) -> LeaderboardApiClient:
        return self.get_new_leaderboard_client()

    def get_new_leaderboard_client(self) -> HostedAlphaLeaderboardApiClient:
        if self._new_leaderboard_client is None:
            self._new_leaderboard_client = HostedAlphaLeaderboardApiClient(backend_api_client=self)
        return self._new_leaderboard_client

    @legacy_with_api_exceptions_handler
    def _create_authenticator(self, api_token, ssl_verify, proxies, backend_client):
        return NeptuneAuthenticator(api_token, backend_client, ssl_verify, proxies)

    def _verify_version(self):
        parsed_version = version.parse(self.client_lib_version)

        if self._client_config.min_compatible_version and self._client_config.min_compatible_version > parsed_version:
            click.echo(
                "ERROR: Minimal supported client version is {} (installed: {}). Please upgrade neptune-client".format(
                    self._client_config.min_compatible_version, self.client_lib_version
                ),
                sys.stderr,
            )
            raise UnsupportedClientVersion(
                self.client_lib_version,
                self._client_config.min_compatible_version,
                self._client_config.max_compatible_version,
            )
        if self._client_config.max_compatible_version and self._client_config.max_compatible_version < parsed_version:
            click.echo(
                "ERROR: Maximal supported client version is {} (installed: {}). Please downgrade neptune-client".format(
                    self._client_config.max_compatible_version, self.client_lib_version
                ),
                sys.stderr,
            )
            raise UnsupportedClientVersion(
                self.client_lib_version,
                self._client_config.min_compatible_version,
                self._client_config.max_compatible_version,
            )
        if self._client_config.min_recommended_version and self._client_config.min_recommended_version > parsed_version:
            click.echo(
                "WARNING: We recommend an upgrade to a new version of neptune-client - {} (installed - {}).".format(
                    self._client_config.min_recommended_version, self.client_lib_version
                ),
                sys.stderr,
            )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "DEFAULT_FLUSH_PERIOD",
    "DEFAULT_NAME",
    "OFFLINE_PROJECT_QUALIFIED_NAME",
    "ASYNC_LAG_THRESHOLD",
    "ASYNC_NO_PROGRESS_THRESHOLD",
    "DEFAULT_STOP_TIMEOUT",
]

DEFAULT_FLUSH_PERIOD = 5
DEFAULT_NAME = "Untitled"
OFFLINE_PROJECT_QUALIFIED_NAME = "offline/project-placeholder"
ASYNC_LAG_THRESHOLD = 1800.0
ASYNC_NO_PROGRESS_THRESHOLD = 300.0
DEFAULT_STOP_TIMEOUT = 60.0



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "InternalClientError",
    "NeptuneException",
    "NeptuneInvalidApiTokenException",
    "NeptuneApiException",
    "MetadataInconsistency",
    "MissingFieldException",
    "TypeDoesNotSupportAttributeException",
    "MalformedOperation",
    "FileNotFound",
    "FileUploadError",
    "FileSetUploadError",
    "FileSetNotFound",
    "ClientHttpError",
    "MetadataContainerNotFound",
    "ProjectNotFound",
    "RunNotFound",
    "ModelNotFound",
    "ModelVersionNotFound",
    "ExceptionWithProjectsWorkspacesListing",
    "ContainerUUIDNotFound",
    "RunUUIDNotFound",
    "ProjectNotFoundWithSuggestions",
    "AmbiguousProjectName",
    "NeptuneMissingProjectNameException",
    "InactiveContainerException",
    "InactiveRunException",
    "InactiveModelException",
    "InactiveModelVersionException",
    "InactiveProjectException",
    "NeptuneMissingApiTokenException",
    "CannotSynchronizeOfflineRunsWithoutProject",
    "NeedExistingExperimentForReadOnlyMode",
    "NeedExistingRunForReadOnlyMode",
    "NeedExistingModelForReadOnlyMode",
    "NeedExistingModelVersionForReadOnlyMode",
    "NeptuneParametersCollision",
    "NeptuneWrongInitParametersException",
    "NeptuneRunResumeAndCustomIdCollision",
    "NeptuneClientUpgradeRequiredError",
    "NeptuneMissingRequiredInitParameter",
    "CannotResolveHostname",
    "NeptuneSSLVerificationError",
    "NeptuneConnectionLostException",
    "InternalServerError",
    "Unauthorized",
    "Forbidden",
    "NeptuneOfflineModeException",
    "NeptuneOfflineModeFetchException",
    "NeptuneOfflineModeChangeStageException",
    "NeptuneProtectedPathException",
    "NeptuneCannotChangeStageManually",
    "OperationNotSupported",
    "NeptuneLegacyProjectException",
    "NeptuneIntegrationNotInstalledException",
    "NeptuneLimitExceedException",
    "NeptuneFieldCountLimitExceedException",
    "NeptuneStorageLimitException",
    "FetchAttributeNotFoundException",
    "ArtifactNotFoundException",
    "PlotlyIncompatibilityException",
    "NeptunePossibleLegacyUsageException",
    "NeptuneLegacyIncompatibilityException",
    "NeptuneUnhandledArtifactSchemeException",
    "NeptuneUnhandledArtifactTypeException",
    "NeptuneLocalStorageAccessException",
    "NeptuneRemoteStorageCredentialsException",
    "NeptuneRemoteStorageAccessException",
    "ArtifactUploadingError",
    "NeptuneUnsupportedArtifactFunctionalityException",
    "NeptuneEmptyLocationException",
    "NeptuneFeatureNotAvailableException",
    "NeptuneObjectCreationConflict",
    "NeptuneModelKeyAlreadyExistsError",
    "NeptuneSynchronizationAlreadyStoppedException",
    "StreamAlreadyUsedException",
    "NeptuneUserApiInputException",
]

from typing import (
    List,
    Optional,
    Union,
)
from urllib.parse import urlparse

from packaging.version import Version

from neptune.common.envs import API_TOKEN_ENV_NAME

# Backward compatibility import
from neptune.common.exceptions import (
    STYLES,
    ClientHttpError,
    Forbidden,
    InternalClientError,
    InternalServerError,
    NeptuneApiException,
    NeptuneConnectionLostException,
    NeptuneException,
    NeptuneInvalidApiTokenException,
    NeptuneSSLVerificationError,
    Unauthorized,
)
from neptune.envs import (
    CUSTOM_RUN_ID_ENV_NAME,
    PROJECT_ENV_NAME,
)
from neptune.internal.backends.api_model import (
    Project,
    Workspace,
)
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import QualifiedName
from neptune.internal.utils import replace_patch_version
from neptune.internal.utils.paths import path_to_str


class MetadataInconsistency(NeptuneException):
    pass


class MissingFieldException(NeptuneException, AttributeError, KeyError):
    """Raised when get-like action is called on `Handler`, instead of on `Attribute`."""

    def __init__(self, field_path):
        message = """
{h1}
----MissingFieldException-------------------------------------------------------
{end}
The field "{field_path}" was not found.

There are two possible reasons:
    - There is a typo in the path. Double-check your code for typos.
    - You are fetching a field that another process created, but the local representation is not synchronized.
    If you are sending metadata from multiple processes at the same time, synchronize the local representation before fetching values:
        {python}run.sync(){end}

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""  # noqa: E501
        self._msg = message.format(field_path=field_path, **STYLES)
        super().__init__(self._msg)

    def __str__(self):
        # required because of overriden `__str__` in `KeyError`
        return self._msg


class TypeDoesNotSupportAttributeException(NeptuneException, AttributeError):
    def __init__(self, type_, attribute):
        message = """
{h1}
----TypeDoesNotSupportAttributeException----------------------------------------
{end}
{type} has no attribute {attribute}.

{correct}Need help?{end}-> https://docs.neptune.ai/help/error_type_does_not_support_attribute/
"""
        self._msg = message.format(type=type_, attribute=attribute, **STYLES)
        super().__init__(self._msg)

    def __str__(self):
        # required because of overriden `__str__` in `KeyError`
        return self._msg


class MalformedOperation(NeptuneException):
    pass


class FileNotFound(NeptuneException):
    def __init__(self, file: str):
        super().__init__("File not found: {}".format(file))


class FileUploadError(NeptuneException):
    def __init__(self, filename: str, msg: str):
        super().__init__("Cannot upload file {}: {}".format(filename, msg))


class FileSetUploadError(NeptuneException):
    def __init__(self, globs: List[str], msg: str):
        super().__init__("Cannot upload file set {}: {}".format(globs, msg))


class MetadataContainerNotFound(NeptuneException):
    container_id: str
    container_type: ContainerType

    def __init__(self, container_id: str, container_type: Optional[ContainerType]):
        self.container_id = container_id
        self.container_type = container_type
        container_type_str = container_type.value.capitalize() if container_type else "object"
        super().__init__("{} {} not found.".format(container_type_str, container_id))

    @classmethod
    def of_container_type(cls, container_type: Optional[ContainerType], container_id: str):
        if container_type is None:
            return MetadataContainerNotFound(container_id=container_id, container_type=None)
        elif container_type == ContainerType.PROJECT:
            return ProjectNotFound(project_id=container_id)
        elif container_type == ContainerType.RUN:
            return RunNotFound(run_id=container_id)
        elif container_type == ContainerType.MODEL:
            return ModelNotFound(model_id=container_id)
        elif container_type == ContainerType.MODEL_VERSION:
            return ModelVersionNotFound(model_version_id=container_id)
        else:
            raise InternalClientError(f"Unexpected ContainerType: {container_type}")


class ProjectNotFound(MetadataContainerNotFound):
    def __init__(self, project_id: str):
        super().__init__(container_id=project_id, container_type=ContainerType.PROJECT)


class RunNotFound(MetadataContainerNotFound):
    def __init__(self, run_id: str):
        super().__init__(container_id=run_id, container_type=ContainerType.RUN)


class ModelNotFound(MetadataContainerNotFound):
    def __init__(self, model_id: str):
        super().__init__(container_id=model_id, container_type=ContainerType.MODEL)


class ModelVersionNotFound(MetadataContainerNotFound):
    def __init__(self, model_version_id: str):
        super().__init__(container_id=model_version_id, container_type=ContainerType.MODEL_VERSION)


class ExceptionWithProjectsWorkspacesListing(NeptuneException):
    def __init__(
        self,
        message: str,
        available_projects: List[Project] = (),
        available_workspaces: List[Workspace] = (),
        **kwargs,
    ):
        available_projects_message = """
Did you mean any of these?
{projects}
"""

        available_workspaces_message = """
You can check all of your projects on the Projects page:
{workspaces_urls}
"""

        projects_formated_list = "\n".join(
            map(
                lambda project: f"    - {project.workspace}/{project.name}",
                available_projects,
            )
        )

        workspaces_formated_list = "\n".join(
            map(
                lambda workspace: f"    - https://app.neptune.ai/{workspace.name}/-/projects",
                available_workspaces,
            )
        )

        super().__init__(
            message.format(
                available_projects_message=available_projects_message.format(projects=projects_formated_list)
                if available_projects
                else "",
                available_workspaces_message=available_workspaces_message.format(
                    workspaces_urls=workspaces_formated_list
                )
                if available_workspaces
                else "",
                **STYLES,
                **kwargs,
            )
        )


class ContainerUUIDNotFound(NeptuneException):
    container_id: str
    container_type: ContainerType

    def __init__(self, container_id: str, container_type: ContainerType):
        self.container_id = container_id
        self.container_type = container_type
        super().__init__(
            "{} with ID {} not found. It may have been deleted. "
            "You can use the 'neptune clear' command to delete junk objects from local storage.".format(
                container_type.value.capitalize(), container_id
            )
        )


# for backward compatibility
RunUUIDNotFound = ContainerUUIDNotFound


class ProjectNotFoundWithSuggestions(ExceptionWithProjectsWorkspacesListing, ProjectNotFound):
    def __init__(
        self,
        project_id: QualifiedName,
        available_projects: List[Project] = (),
        available_workspaces: List[Workspace] = (),
    ):
        message = """
{h1}
----NeptuneProjectNotFoundException------------------------------------
{end}
We couldn't find project {fail}"{project}"{end}.
{available_projects_message}{available_workspaces_message}
You may want to check the following docs page:
    - https://docs.neptune.ai/setup/creating_project/

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(
            message=message,
            available_projects=available_projects,
            available_workspaces=available_workspaces,
            project=project_id,
        )


class AmbiguousProjectName(ExceptionWithProjectsWorkspacesListing):
    def __init__(self, project_id: str, available_projects: List[Project] = ()):
        message = """
{h1}
----NeptuneProjectNameCollisionException------------------------------------
{end}
Cannot resolve project {fail}"{project}"{end}. Name is ambiguous.
{available_projects_message}
You may also want to check the following docs pages:
    - https://docs.neptune.ai/setup/creating_project/

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message=message, available_projects=available_projects, project=project_id)


class NeptuneMissingProjectNameException(ExceptionWithProjectsWorkspacesListing):
    def __init__(
        self,
        available_projects: List[Project] = (),
        available_workspaces: List[Workspace] = (),
    ):
        message = """
{h1}
----NeptuneMissingProjectNameException----------------------------------------
{end}
The Neptune client couldn't find your project name.
{available_projects_message}{available_workspaces_message}
There are two options to add it:
    - specify it in your code
    - set an environment variable in your operating system.

{h2}CODE{end}
Pass it to the {bold}init_run(){end} function via the {bold}project{end} argument:
    {python}neptune.init_run(project='WORKSPACE_NAME/PROJECT_NAME'){end}

{h2}ENVIRONMENT VARIABLE{end}
or export or set an environment variable depending on your operating system:

    {correct}Linux/Unix{end}
    In your terminal run:
       {bash}export {env_project}=WORKSPACE_NAME/PROJECT_NAME{end}

    {correct}Windows{end}
    In your CMD run:
       {bash}set {env_project}=WORKSPACE_NAME/PROJECT_NAME{end}

and skip the {bold}project{end} argument of the {bold}init_run(){end} function:
    {python}neptune.init_run(){end}

You may also want to check the following docs pages:
    - https://docs.neptune.ai/setup/creating_project/

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(
            message=message,
            available_projects=available_projects,
            available_workspaces=available_workspaces,
            env_project=PROJECT_ENV_NAME,
        )


class InactiveContainerException(NeptuneException):
    resume_info: str

    def __init__(self, container_type: ContainerType, label: str):
        message = """
{h1}
----{cls}----------------------------------------
{end}
It seems you are trying to log metadata to (or fetch it from) a {container_type} that was stopped ({label}).

Here's what you can do:{resume_info}

You may also want to check the following docs pages:
    - https://docs.neptune.ai/logging/to_existing_object/
    - https://docs.neptune.ai/usage/querying_metadata/

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(
            message.format(
                cls=self.__class__.__name__,
                label=label,
                container_type=container_type.value,
                resume_info=self.resume_info,
                **STYLES,
            )
        )


class InactiveRunException(InactiveContainerException):
    resume_info = """
    - Resume the run to continue logging to it:
    https://docs.neptune.ai/logging/to_existing_object/
    - Don't invoke `stop()` on a run that you want to access. If you want to stop monitoring only,
    you can resume a run in read-only mode:
    https://docs.neptune.ai/api/connection_modes/#read-only-mode"""

    def __init__(self, label: str):
        super().__init__(label=label, container_type=ContainerType.RUN)


class InactiveModelException(InactiveContainerException):
    resume_info = """
    - Resume the model to continue logging to it:
    https://docs.neptune.ai/api/neptune/#init_model
    - Don't invoke `stop()` on a model that you want to access. If you want to stop monitoring only,
    you can resume a model in read-only mode:
    https://docs.neptune.ai/api/connection_modes/#read-only-mode"""

    def __init__(self, label: str):
        super().__init__(label=label, container_type=ContainerType.MODEL)


class InactiveModelVersionException(InactiveContainerException):
    resume_info = """
    - Resume the model version to continue logging to it:
    https://docs.neptune.ai/api/neptune/#init_model_version
    - Don't invoke `stop()` on a model version that you want to access. If you want to stop monitoring only,
    you can resume a model version in read-only mode:
    https://docs.neptune.ai/api/connection_modes/#read-only-mode"""

    def __init__(self, label: str):
        super().__init__(label=label, container_type=ContainerType.MODEL_VERSION)


class InactiveProjectException(InactiveContainerException):
    resume_info = """
    - Resume the connection to the project to continue logging to it:
    https://docs.neptune.ai/api/neptune/#init_project
    - Don't invoke `stop()` on a project that you want to access."""

    def __init__(self, label: str):
        super().__init__(label=label, container_type=ContainerType.PROJECT)


class NeptuneMissingApiTokenException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneMissingApiTokenException-------------------------------------------
{end}
The Neptune client couldn't find your API token.

You can get it here:
    - https://app.neptune.ai/get_my_api_token

There are two options to add it:
    - specify it in your code
    - set an environment variable in your operating system.

{h2}CODE{end}
Pass the token to the {bold}init_run(){end} function via the {bold}api_token{end} argument:
    {python}neptune.init_run(project='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}

{h2}ENVIRONMENT VARIABLE{end} {correct}(Recommended option){end}
or export or set an environment variable depending on your operating system:

    {correct}Linux/Unix{end}
    In your terminal run:
        {bash}export {env_api_token}="YOUR_API_TOKEN"{end}

    {correct}Windows{end}
    In your CMD run:
        {bash}set {env_api_token}="YOUR_API_TOKEN"{end}

and skip the {bold}api_token{end} argument of the {bold}init_run(){end} function:
    {python}neptune.init_run(project='WORKSPACE_NAME/PROJECT_NAME'){end}

You may also want to check the following docs pages:
    - https://docs.neptune.ai/setup/setting_api_token/

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(env_api_token=API_TOKEN_ENV_NAME, **STYLES))


class CannotSynchronizeOfflineRunsWithoutProject(NeptuneException):
    def __init__(self):
        super().__init__("Cannot synchronize offline runs without a project.")


class NeedExistingExperimentForReadOnlyMode(NeptuneException):
    container_type: ContainerType
    callback_name: str

    def __init__(self, container_type: ContainerType, callback_name: str):
        message = """
{h1}
----{class_name}-----------------------------------------
{end}
Read-only mode can be used only with an existing {container_type}.

Pass the ID of a {container_type} to the {python}with_id{end} argument of {python}{callback_name}{end}
when using {python}mode="read-only"{end}.

You may also want to check the following docs pages:
    - https://docs.neptune.ai/usage/resume_run/
    - https://docs.neptune.ai/api/connection_modes/#read-only-mode

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        self.container_type = container_type
        self.callback_name = callback_name
        super().__init__(
            message.format(
                class_name=type(self).__name__,
                container_type=self.container_type.value,
                callback_name=self.callback_name,
                **STYLES,
            )
        )


class NeedExistingRunForReadOnlyMode(NeedExistingExperimentForReadOnlyMode):
    def __init__(self):
        super().__init__(container_type=ContainerType.RUN, callback_name="neptune.init_run")


class NeedExistingModelForReadOnlyMode(NeedExistingExperimentForReadOnlyMode):
    def __init__(self):
        super().__init__(container_type=ContainerType.MODEL, callback_name="neptune.init_model")


class NeedExistingModelVersionForReadOnlyMode(NeedExistingExperimentForReadOnlyMode):
    def __init__(self):
        super().__init__(
            container_type=ContainerType.MODEL_VERSION,
            callback_name="neptune.init_model_version",
        )


class NeptuneParametersCollision(NeptuneException):
    def __init__(self, parameter1, parameter2, method_name):
        self.parameter1 = parameter1
        self.parameter2 = parameter2
        self.method_name = method_name
        message = """
{h1}
----NeptuneParametersCollision-----------------------------------------
{end}
The {python}{parameter1}{end} and {python}{parameter2}{end} parameters of the {python}{method_name}(){end} method are mutually exclusive.

You may also want to check the following docs page:
    - https://docs.neptune.ai/api/universal/#initialization-methods

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""  # noqa: E501
        super().__init__(
            message.format(
                parameter1=parameter1,
                parameter2=parameter2,
                method_name=method_name,
                **STYLES,
            )
        )


class NeptuneWrongInitParametersException(NeptuneException):
    pass


class NeptuneRunResumeAndCustomIdCollision(NeptuneWrongInitParametersException):
    def __init__(self):
        message = """
{h1}
----NeptuneRunResumeAndCustomIdCollision-----------------------------------------
{end}
It's not possible to use {python}custom_run_id{end} while resuming a run.

The {python}run{end} and {python}custom_run_id{end} parameters of the {python}init_run(){end} method are mutually exclusive.
Make sure you have no {bash}{custom_id_env}{end} environment variable set
and no value is explicitly passed to the `custom_run_id` argument when you are resuming a run.

You may also want to check the following docs page:
    - https://docs.neptune.ai/logging/to_existing_object/
    - https://docs.neptune.ai/logging/custom_run_id/

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""  # noqa: E501
        super().__init__(message.format(custom_id_env=CUSTOM_RUN_ID_ENV_NAME, **STYLES))


class NeptuneClientUpgradeRequiredError(NeptuneException):
    def __init__(
        self,
        version: Union[Version, str],
        min_version: Optional[Union[Version, str]] = None,
        max_version: Optional[Union[Version, str]] = None,
    ):
        current_version = str(version)
        required_version = "==" + replace_patch_version(str(max_version)) if max_version else ">=" + str(min_version)
        message = """
{h1}
----NeptuneClientUpgradeRequiredError-------------------------------------------------------------
{end}
Your version of the Neptune client library ({current_version}) is no longer supported by the Neptune
 server. The minimum required version is {required_version}.

In order to update the Neptune client library, run the following command in your terminal:
    {bash}pip install -U neptune{end}
Or if you are using Conda, run the following instead:
    {bash}conda update -c conda-forge neptune-client{end}

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(
            message.format(
                current_version=current_version,
                required_version=required_version,
                **STYLES,
            )
        )


class NeptuneMissingRequiredInitParameter(NeptuneWrongInitParametersException):
    def __init__(
        self,
        called_function: str,
        parameter_name: str,
    ):
        message = """
{h1}
----NeptuneMissingRequiredInitParameter---------------------------------------
{end}
{python}neptune.{called_function}(){end} invocation was missing {python}{parameter_name}{end}.
If you want to create a new object using {python}{called_function}{end}, {python}{parameter_name}{end} is required:
https://docs.neptune.ai/api/neptune#{called_function}

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(
            message.format(
                called_function=called_function,
                parameter_name=parameter_name,
                **STYLES,
            )
        )


class CannotResolveHostname(NeptuneException):
    def __init__(self, host):
        message = """
{h1}
----CannotResolveHostname-----------------------------------------------------------------------
{end}
The Neptune client library was not able to resolve hostname {underline}{host}{end}.

What should I do?
    - Check if your computer is connected to the internet.
    - Check if your computer is supposed to be using a proxy to access the internet.
      If so, you may want to use the {python}proxies{end} parameter of the {python}init_run(){end} function.
      See https://docs.neptune.ai/api/universal/#proxies
      and https://requests.readthedocs.io/en/latest/user/advanced/#proxies
    - Check the status of Neptune services: https://status.neptune.ai/

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(host=host, **STYLES))


class NeptuneOfflineModeException(NeptuneException):
    pass


class NeptuneOfflineModeFetchException(NeptuneOfflineModeException):
    def __init__(self):
        message = """
{h1}
----NeptuneOfflineModeFetchException---------------------------------------------------
{end}
It seems you are trying to fetch data from the server while working in offline mode.
You need to work in a non-offline connection mode to fetch data from the server.

You can set the connection mode when creating a new run:
    {python}run = neptune.init_run(mode="async"){end}

You may also want to check the following docs page:
    - https://docs.neptune.ai/api/connection_modes

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(**STYLES))


class NeptuneOfflineModeChangeStageException(NeptuneOfflineModeException):
    def __init__(self):
        message = """
{h1}
----NeptuneOfflineModeChangeStageException---------------------------------------
{end}
You cannot change the stage of the model version while in offline mode.
"""
        super().__init__(message.format(**STYLES))


class NeptuneProtectedPathException(NeptuneException):
    extra_info = ""

    def __init__(self, path: str):
        message = """
{h1}
----NeptuneProtectedPathException----------------------------------------------
{end}
Field {path} cannot be changed directly.
{extra_info}

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        self._path = path
        super().__init__(
            message.format(
                path=path,
                extra_info=self.extra_info.format(**STYLES),
                **STYLES,
            )
        )


class NeptuneCannotChangeStageManually(NeptuneProtectedPathException):
    extra_info = """
If you want to change the stage of the model version,
use the {python}.change_stage(){end} method:
    {python}model_version.change_stage("staging"){end}"""


class OperationNotSupported(NeptuneException):
    def __init__(self, message: str):
        super().__init__(f"Operation not supported: {message}")


class NeptuneLegacyProjectException(NeptuneException):
    def __init__(self, project: QualifiedName):
        message = """
{h1}
----NeptuneLegacyProjectException---------------------------------------------------------
{end}
Your project "{project}" has not been migrated to the new structure yet.
Unfortunately the neptune.new Python API is incompatible with projects using the old structure,
so please use legacy neptune Python API.
Don't worry - we are working hard on migrating all the projects and you will be able to use the neptune.new API soon.

You can find documentation for the legacy neptune Python API here:
    - https://docs-legacy.neptune.ai/index.html

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(project=project, **STYLES))


class NeptuneIntegrationNotInstalledException(NeptuneException):
    def __init__(self, integration_package_name, framework_name):
        message = """
{h1}
----NeptuneIntegrationNotInstalledException-----------------------------------------
{end}
Looks like integration {integration_package_name} wasn't installed.
To install, run:
    {bash}pip install {integration_package_name}{end}
Or:
    {bash}pip install "neptune[{framework_name}]"{end}

You may also want to check the following docs page:
    - https://docs.neptune.ai/integrations

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(
            message.format(
                integration_package_name=integration_package_name,
                framework_name=framework_name,
                **STYLES,
            )
        )


class NeptuneLimitExceedException(NeptuneException):
    def __init__(self, reason: str):
        message = """
{h1}
----NeptuneLimitExceedException---------------------------------------------------------------------------------------
{end}
{reason}

Data synchronization was interrupted because you're out of storage space (or logging hours). If you're using Neptune in
asynchronous mode (default), the data that couldn't be uploaded is safely stored on the disk.
You can still fetch and delete data from your project(s).
How can I free up space in Neptune?
    - Contact your workspace admin about possible project or workspace storage limits.
    - Upgrade your subscription with more storage: https://app.neptune.ai/-/subscription
    - Delete run or model metadata that you don't need.
How do I upload my offline metadata to Neptune?
    You can upload the data stored on-disk with the following command (replace the workspace and project with your own):
        {bash}neptune sync -p workspace-name/project-name{end}
If your subscription is based on logging hours, you may also get this error if you've run out of monitoring time.
    To continue logging, you need to top up your logging hours or wait for your new monthly package.
Learn more in the docs:
    - https://docs.neptune.ai/help/error_limit_exceeded/
    - https://docs.neptune.ai/help/workspace_read_only/
    - https://docs.neptune.ai/api/connection_modes/
    - https://docs.neptune.ai/usage/deleting_data/
    - https://docs.neptune.ai/api/neptune_sync/
    - https://docs.neptune.ai/help/storage_tips/
{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(**STYLES, reason=reason))


class NeptuneFieldCountLimitExceedException(NeptuneException):
    def __init__(self, limit: int, container_type: str, identifier: str):
        message = """
{h1}
----NeptuneFieldCountLimitExceedException---------------------------------------------------------------------------------------
{end}
There are too many fields (more than {limit}) in the {identifier} {container_type}.
We have stopped the synchronization to the Neptune server and stored the data locally.

To continue uploading the metadata:

    1. Delete some excess fields from {identifier}.

       You can delete fields or namespaces with the "del" command.
       For example, to delete the "training/checkpoints" namespace:

       {python}del run["training/checkpoints"]{end}

    2. Once you're done, synchronize the data manually with the following command:

       {bash}neptune sync -p project_name{end}

For more details, see https://docs.neptune.ai/usage/best_practices
"""  # noqa: E501
        super().__init__(
            message.format(
                **STYLES,
                limit=limit,
                container_type=container_type,
                identifier=identifier,
            )
        )


class NeptuneStorageLimitException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneStorageLimitException---------------------------------------------------------------------------------------
{end}
You exceeded the storage limit of the workspace. It's not possible to upload new data, but you can still fetch and delete data.
If you are using asynchronous (default) connection mode, Neptune automatically switched to offline mode
and your data is being stored safely on the disk. You can upload it later using the Neptune Command Line Interface tool:
    {bash}neptune sync -p project_name{end}
What should I do?
    - Go to your projects and remove runs or model metadata you don't need
    - ... or update your subscription plan here: https://app.neptune.ai/-/subscription
You may also want to check the following docs page:
    - https://docs.neptune.ai/api/connection_modes
{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""  # noqa: E501
        super().__init__(message.format(**STYLES))


class FetchAttributeNotFoundException(MetadataInconsistency):
    def __init__(self, attribute_path: str):
        message = """
{h1}
----MetadataInconsistency----------------------------------------------------------------------
{end}
The field {python}{attribute_path}{end} was not found.

Remember that in the asynchronous (default) connection mode, data is synchronized
with the Neptune servers in the background. The data may have not reached
the servers before it was fetched. Before fetching the data, you can force
wait for all the requests sent by invoking:

    {python}run.wait(){end}

Remember that each use of {python}wait{end} introduces a delay in code execution.

You may also want to check the following docs page:
    - https://docs.neptune.ai/api/connection_modes

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help.html
"""
        super().__init__(message.format(attribute_path=attribute_path, **STYLES))


class ArtifactNotFoundException(MetadataInconsistency):
    def __init__(self, artifact_hash: str):
        message = """
{h1}
----MetadataInconsistency----------------------------------------------------------------------
{end}
Artifact with hash {python}{artifact_hash}{end} was not found.

Remember that in the asynchronous (default) connection mode, data is synchronized
with the Neptune servers in the background. The data may have not reached
the servers before it was fetched. Before fetching the data, you can force
wait for all the requests sent by invoking:

    {python}run.wait(){end}

Remember that each use of {python}wait{end} introduces a delay in code execution.

You may also want to check the following docs page:
    - https://docs.neptune.ai/api/connection_modes

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help.html
"""
        super().__init__(message.format(artifact_hash=artifact_hash, **STYLES))


class PlotlyIncompatibilityException(Exception):
    def __init__(self, matplotlib_version, plotly_version, details):
        super().__init__(
            "Unable to convert plotly figure to matplotlib format. "
            "Your matplotlib ({}) and plotlib ({}) versions are not compatible. "
            "{}".format(matplotlib_version, plotly_version, details)
        )


class NeptunePossibleLegacyUsageException(NeptuneWrongInitParametersException):
    def __init__(self):
        message = """
{h1}
----NeptunePossibleLegacyUsageException----------------------------------------------------------------
{end}
It seems you are trying to use the legacy API, but you imported the new one.

Simply update your import statement to:
    {python}import neptune{end}

You may want to check the legacy API docs:
    - https://docs-legacy.neptune.ai

If you want to update your code with the new API, we prepared a handy migration guide:
    - https://docs.neptune.ai/about/legacy/#migrating-to-neptunenew

You can read more about neptune.new in the release blog post:
    - https://neptune.ai/blog/neptune-new

You may also want to check the following docs page:
    - https://docs-legacy.neptune.ai/getting-started/integrate-neptune-into-your-codebase.html

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(**STYLES))


class NeptuneLegacyIncompatibilityException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneLegacyIncompatibilityException----------------------------------------
{end}
It seems you are passing the legacy Experiment object, when a Run object is expected.

What can I do?
    - Updating your code to the new Python API requires few changes, but to help you with this process we prepared a handy migration guide:
    https://docs.neptune.ai/about/legacy/#migrating-to-neptunenew
    - You can read more about neptune.new in the release blog post:
    https://neptune.ai/blog/neptune-new

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""  # noqa: E501
        super().__init__(message.format(**STYLES))


class NeptuneUnhandledArtifactSchemeException(NeptuneException):
    def __init__(self, path: str):
        scheme = urlparse(path).scheme
        message = """
{h1}
----NeptuneUnhandledArtifactProtocolException------------------------------------
{end}
You have used a Neptune Artifact to track a file with a scheme unhandled by this client ({scheme}).
Problematic path: {path}

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(scheme=scheme, path=path, **STYLES))


class NeptuneUnhandledArtifactTypeException(NeptuneException):
    def __init__(self, type_str: str):
        message = """
{h1}
----NeptuneUnhandledArtifactTypeException----------------------------------------
{end}
A Neptune Artifact you're listing is tracking a file type unhandled by this client ({type_str}).

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(type_str=type_str, **STYLES))


class NeptuneLocalStorageAccessException(NeptuneException):
    def __init__(self, path, expected_description):
        message = """
{h1}
----NeptuneLocalStorageAccessException-------------------------------------
{end}
Neptune had a problem processing "{path}". It expects it to be {expected_description}.

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(path=path, expected_description=expected_description, **STYLES))


class NeptuneRemoteStorageCredentialsException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneRemoteStorageCredentialsException-------------------------------------
{end}
Neptune could not find suitable credentials for remote storage of a Neptune Artifact you're listing.

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(**STYLES))


class NeptuneRemoteStorageAccessException(NeptuneException):
    def __init__(self, location: str):
        message = """
{h1}
----NeptuneRemoteStorageAccessException------------------------------------------
{end}
Neptune could not access an object ({location}) from remote storage of a Neptune Artifact you're listing.

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(location=location, **STYLES))


class ArtifactUploadingError(NeptuneException):
    def __init__(self, msg: str):
        super().__init__("Cannot upload artifact: {}".format(msg))


class NeptuneUnsupportedArtifactFunctionalityException(NeptuneException):
    def __init__(self, functionality_info: str):
        message = """
{h1}
----NeptuneUnsupportedArtifactFunctionality-------------------------------------
{end}
It seems you are using Neptune Artifacts functionality that is currently not supported.

{functionality_info}

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(functionality_info=functionality_info, **STYLES))


class NeptuneEmptyLocationException(NeptuneException):
    def __init__(self, location: str, namespace: str):
        message = """
{h1}
----NeptuneEmptyLocationException----------------------------------------------
{end}
Neptune could not find files in the requested location ({location}) during the creation of an Artifact in "{namespace}".

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(location=location, namespace=namespace, **STYLES))


class NeptuneFeatureNotAvailableException(NeptuneException):
    def __init__(self, missing_feature):
        message = """
{h1}
----NeptuneFeatureNotAvailableException----------------------------------------------
{end}
The following feature is not yet supported by the Neptune instance you are using:
{missing_feature}

An update of the Neptune instance is required in order to use it. Please contact your local Neptune administrator
or Neptune support directly (support@neptune.ai) about the upcoming updates.

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        self.message = message.format(missing_feature=missing_feature, **STYLES)
        super().__init__(message)


class NeptuneObjectCreationConflict(NeptuneException):
    pass


class NeptuneModelKeyAlreadyExistsError(NeptuneObjectCreationConflict):
    def __init__(self, model_key, models_tab_url):
        message = """
{h1}
----NeptuneModelKeyAlreadyExistsError---------------------------------------------------
{end}
A model with the provided key ({model_key}) already exists in this project. A model key has to be unique
within the project.

You can check all of your models in the project on the Models page:
{models_tab_url}

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(model_key=model_key, models_tab_url=models_tab_url, **STYLES))


class NeptuneSynchronizationAlreadyStoppedException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneSynchronizationAlreadyStopped---------------------------------------------------
{end}
The synchronization thread had stopped before Neptune could finish uploading the logged metadata.
Your data is stored locally, but you'll need to finish the synchronization manually.
To synchronize with the Neptune servers, enter the following on your command line:

    {bash}neptune sync{end}

For details, see https://docs.neptune.ai/api/neptune_sync/

If the synchronization fails, you may want to check your connection and ensure that you're
within limits by going to your Neptune project settings -> Usage.
If the issue persists, our support is happy to help.

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(**STYLES))


class StreamAlreadyUsedException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----StreamAlreadyUsedException---------------------------------------------------
{end}
A File object created with File.from_stream() has already been logged.
You can only log content from the same stream once.

For more, see https://docs.neptune.ai/api/field_types/#from_stream

{correct}Need help?{end}-> https://docs.neptune.ai/getting_help
"""
        super().__init__(message.format(**STYLES))


class NeptuneUserApiInputException(NeptuneException):
    def __init__(self, message):
        super().__init__(message)


class FileSetNotFound(NeptuneException):
    def __init__(self, attribute: str, path: str):
        message = """
        {h1}
        ----MetadataInconsistency----------------------------------------------------------------------
        {end}
        Attribute {python}{path}{end} was not found.

        Remember that in the asynchronous (default) connection mode, data is synchronized
        with the Neptune servers in the background. The data may have not reached
        the servers before it was fetched. Before fetching the data, you can force
        wait for all the requests sent by invoking:

            {python}run.wait(){end}

        Remember that each use of {python}wait{end} introduces a delay in code execution.

        You may also want to check the following docs page:
            - https://docs.neptune.ai/api/connection_modes

        {correct}Need help?{end}-> https://docs.neptune.ai/getting_help/
        """
        super().__init__(message.format(path=path_to_str([attribute, path]), **STYLES))



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Namespace", "NamespaceBuilder"]

import argparse
from collections.abc import MutableMapping
from typing import (
    TYPE_CHECKING,
    Any,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Union,
)

from neptune.attributes.attribute import Attribute
from neptune.internal.container_structure import ContainerStructure
from neptune.internal.utils.generic_attribute_mapper import (
    NoValue,
    atomic_attribute_types_map,
)
from neptune.internal.utils.logger import logger
from neptune.internal.utils.paths import (
    parse_path,
    path_to_str,
)
from neptune.types.namespace import Namespace as NamespaceVal

if TYPE_CHECKING:
    from neptune.metadata_containers import MetadataContainer

RunStructure = ContainerStructure  # backwards compatibility


class Namespace(Attribute, MutableMapping):
    def __init__(self, container: "MetadataContainer", path: List[str]):
        Attribute.__init__(self, container, path)
        self._attributes = {}
        self._str_path = path_to_str(path)

    def __setitem__(self, k: str, v: Attribute) -> None:
        if not parse_path(k):
            logger.warning(
                f'Key "{k}" can\'t be used in Namespaces and dicts stored in Neptune. Please use a non-empty key '
                f"instead. The value {v!r} will be dropped.",
            )
            return
        self._attributes[k] = v

    def __delitem__(self, k: str) -> None:
        del self._attributes[k]

    def __getitem__(self, k: str) -> Attribute:
        return self._attributes[k]

    def __len__(self) -> int:
        return len(self._attributes)

    def __iter__(self) -> Iterator[str]:
        yield from self._attributes.__iter__()

    def extend(
        self,
        value: Union[Any, Iterable[Any]],
        *,
        steps: Optional[Collection[float]] = None,
        timestamps: Optional[Collection[float]] = None,
        wait: bool = False,
        **kwargs,
    ) -> None:
        if not isinstance(value, NamespaceVal):
            value = NamespaceVal(value)
        for k, v in value.value.items():
            self._container[f"{self._str_path}/{k}"].extend(v, steps=steps, timestamps=timestamps, wait=wait, **kwargs)

    def to_dict(self) -> Dict[str, Any]:
        result = {}
        for key, value in self._attributes.items():
            if isinstance(value, Namespace):
                result[key] = value.to_dict()
            else:
                result[key] = value
        return result

    def assign(self, value: Union[NamespaceVal, dict, Mapping], *, wait: bool = False):
        if isinstance(value, argparse.Namespace):
            value = NamespaceVal(vars(value))
        elif not isinstance(value, NamespaceVal):
            value = NamespaceVal(value)

        for k, v in value.value.items():
            self._container[f"{self._str_path}/{k}"].assign(v, wait=wait)

    def _collect_atom_values(self, attribute_dict) -> dict:
        result = {}
        for k, v in attribute_dict.items():
            if isinstance(v, dict):
                result[k] = self._collect_atom_values(v)
            else:
                attr_type, attr_value = v
                if attr_type in atomic_attribute_types_map and attr_value is not NoValue:
                    result[k] = v[1]
        return result

    def fetch(self) -> dict:
        attributes = self._backend.fetch_atom_attribute_values(self._container_id, self._container_type, self._path)
        run_struct = ContainerStructure()
        prefix_len = len(self._path)
        for attr_name, attr_type, attr_value in attributes:
            run_struct.set(parse_path(attr_name)[prefix_len:], (attr_type, attr_value))
        return self._collect_atom_values(run_struct.get_structure())


class NamespaceBuilder:
    def __init__(self, container: "MetadataContainer"):
        self._run = container

    def __call__(self, path: List[str]) -> Namespace:
        return Namespace(self._run, path)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "DEFAULT_REQUEST_KWARGS",
    "create_http_client_with_auth",
    "create_backend_client",
    "create_leaderboard_client",
    "create_artifacts_client",
]

import os
import platform
from typing import (
    Dict,
    Tuple,
)

import requests
from bravado.http_client import HttpClient
from bravado.requests_client import RequestsClient

from neptune.common.backends.utils import with_api_exceptions_handler
from neptune.common.oauth import NeptuneAuthenticator
from neptune.envs import NEPTUNE_REQUEST_TIMEOUT
from neptune.exceptions import NeptuneClientUpgradeRequiredError
from neptune.internal.backends.api_model import ClientConfig
from neptune.internal.backends.swagger_client_wrapper import SwaggerClientWrapper
from neptune.internal.backends.utils import (
    NeptuneResponseAdapter,
    build_operation_url,
    cache,
    create_swagger_client,
    update_session_proxies,
    verify_client_version,
    verify_host_resolution,
)
from neptune.internal.credentials import Credentials
from neptune.version import version as neptune_client_version

BACKEND_SWAGGER_PATH = "/api/backend/swagger.json"
LEADERBOARD_SWAGGER_PATH = "/api/leaderboard/swagger.json"
ARTIFACTS_SWAGGER_PATH = "/api/artifacts/swagger.json"

CONNECT_TIMEOUT = 30  # helps detecting internet connection lost
REQUEST_TIMEOUT = int(os.getenv(NEPTUNE_REQUEST_TIMEOUT, "600"))

DEFAULT_REQUEST_KWARGS = {
    "_request_options": {
        "connect_timeout": CONNECT_TIMEOUT,
        "timeout": REQUEST_TIMEOUT,
        "headers": {"X-Neptune-LegacyClient": "false"},
    }
}


def _close_connections_on_fork(session: requests.Session):
    try:
        os.register_at_fork(before=session.close, after_in_child=session.close, after_in_parent=session.close)
    except AttributeError:
        pass


def create_http_client(ssl_verify: bool, proxies: Dict[str, str]) -> RequestsClient:
    http_client = RequestsClient(ssl_verify=ssl_verify, response_adapter_class=NeptuneResponseAdapter)
    http_client.session.verify = ssl_verify

    _close_connections_on_fork(http_client.session)

    update_session_proxies(http_client.session, proxies)

    user_agent = "neptune-client/{lib_version} ({system}, python {python_version})".format(
        lib_version=neptune_client_version,
        system=platform.platform(),
        python_version=platform.python_version(),
    )
    http_client.session.headers.update({"User-Agent": user_agent})

    return http_client


@cache
def _get_token_client(
    credentials: Credentials,
    ssl_verify: bool,
    proxies: Dict[str, str],
    endpoint_url: str = None,
) -> SwaggerClientWrapper:
    config_api_url = credentials.api_url_opt or credentials.token_origin_address
    if proxies is None:
        verify_host_resolution(config_api_url)

    token_http_client = create_http_client(ssl_verify, proxies)

    return SwaggerClientWrapper(
        create_swagger_client(
            build_operation_url(endpoint_url or config_api_url, BACKEND_SWAGGER_PATH),
            token_http_client,
        )
    )


@cache
@with_api_exceptions_handler
def get_client_config(credentials: Credentials, ssl_verify: bool, proxies: Dict[str, str]) -> ClientConfig:
    backend_client = _get_token_client(credentials=credentials, ssl_verify=ssl_verify, proxies=proxies)

    config = (
        backend_client.api.getClientConfig(
            X_Neptune_Api_Token=credentials.api_token,
            alpha="true",
            **DEFAULT_REQUEST_KWARGS,
        )
        .response()
        .result
    )

    client_config = ClientConfig.from_api_response(config)
    if not client_config.version_info:
        raise NeptuneClientUpgradeRequiredError(neptune_client_version, max_version="0.4.111")
    return client_config


@cache
def create_http_client_with_auth(
    credentials: Credentials, ssl_verify: bool, proxies: Dict[str, str]
) -> Tuple[RequestsClient, ClientConfig]:
    client_config = get_client_config(credentials=credentials, ssl_verify=ssl_verify, proxies=proxies)

    config_api_url = credentials.api_url_opt or credentials.token_origin_address

    verify_client_version(client_config, neptune_client_version)

    endpoint_url = None
    if config_api_url != client_config.api_url:
        endpoint_url = build_operation_url(client_config.api_url, BACKEND_SWAGGER_PATH)

    http_client = create_http_client(ssl_verify=ssl_verify, proxies=proxies)
    http_client.authenticator = NeptuneAuthenticator(
        credentials.api_token,
        _get_token_client(
            credentials=credentials,
            ssl_verify=ssl_verify,
            proxies=proxies,
            endpoint_url=endpoint_url,
        ),
        ssl_verify,
        proxies,
    )

    return http_client, client_config


@cache
def create_backend_client(client_config: ClientConfig, http_client: HttpClient) -> SwaggerClientWrapper:
    return SwaggerClientWrapper(
        create_swagger_client(
            build_operation_url(client_config.api_url, BACKEND_SWAGGER_PATH),
            http_client,
        )
    )


@cache
def create_leaderboard_client(client_config: ClientConfig, http_client: HttpClient) -> SwaggerClientWrapper:
    return SwaggerClientWrapper(
        create_swagger_client(
            build_operation_url(client_config.api_url, LEADERBOARD_SWAGGER_PATH),
            http_client,
        )
    )


@cache
def create_artifacts_client(client_config: ClientConfig, http_client: HttpClient) -> SwaggerClientWrapper:
    return SwaggerClientWrapper(
        create_swagger_client(
            build_operation_url(client_config.api_url, ARTIFACTS_SWAGGER_PATH),
            http_client,
        )
    )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["StringifyValue", "extract_if_stringify_value"]

from typing import Any

from neptune.constants import (
    MAX_32_BIT_INT,
    MIN_32_BIT_INT,
)
from neptune.internal.utils.logger import logger


class StringifyValue:
    def __init__(self, value: Any):
        # check if it's an integer outside 32bit range and cast it to float
        if isinstance(value, int) and (value > MAX_32_BIT_INT or value < MIN_32_BIT_INT):
            logger.info(
                "Value '%d' is outside the range of 32-bit integers ('%d' to '%d') and will be logged as float",
                value,
                MIN_32_BIT_INT,
                MAX_32_BIT_INT,
            )
            value = float(value)

        self.__value = value

    @property
    def value(self):
        return self.__value

    def __str__(self):
        return str(self.__value)

    def __repr__(self):
        return repr(self.__value)


def extract_if_stringify_value(val):
    if isinstance(val, StringifyValue):
        return val.value
    return val



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["path_option"]

from pathlib import Path

import click

from neptune.common.exceptions import NeptuneException  # noqa: F401
from neptune.constants import NEPTUNE_DATA_DIRECTORY
from neptune.exceptions import (  # noqa: F401
    CannotSynchronizeOfflineRunsWithoutProject,
    ProjectNotFound,
    RunNotFound,
)
from neptune.internal.backends.api_model import (  # noqa: F401
    ApiExperiment,
    Project,
)
from neptune.internal.backends.neptune_backend import NeptuneBackend  # noqa: F401
from neptune.internal.disk_queue import DiskQueue  # noqa: F401
from neptune.internal.operation import Operation  # noqa: F401


def get_neptune_path(ctx, param, path: str) -> Path:
    # check if path exists and contains a '.neptune' folder
    path = Path(path)
    if (path / NEPTUNE_DATA_DIRECTORY).is_dir():
        return path / NEPTUNE_DATA_DIRECTORY
    elif path.name == NEPTUNE_DATA_DIRECTORY and path.is_dir():
        return path
    else:
        raise click.BadParameter("Path {} does not contain a '{}' folder.".format(path, NEPTUNE_DATA_DIRECTORY))


path_option = click.option(
    "--path",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    default=Path.cwd(),
    callback=get_neptune_path,
    metavar="<location>",
    help="path to a directory containing a '.neptune' folder with stored objects",
)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["OperationStorage", "get_container_dir"]

import os
import shutil
from pathlib import Path
from typing import Optional

from neptune.constants import NEPTUNE_DATA_DIRECTORY
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import UniqueId
from neptune.internal.utils.logger import logger


def get_container_dir(
    type_dir: str, container_id: UniqueId, container_type: ContainerType, process_path: Optional[str] = None
) -> Path:
    neptune_data_dir = os.getenv("NEPTUNE_DATA_DIRECTORY", NEPTUNE_DATA_DIRECTORY)
    container_dir = Path(f"{neptune_data_dir}/{type_dir}/{container_type.create_dir_name(container_id)}")
    if process_path:
        container_dir /= Path(process_path)

    return container_dir


class OperationStorage:
    UPLOAD_PATH: str = "upload_path"

    def __init__(self, data_path: Path):
        self._data_path = data_path.resolve()

        # initialize directory
        os.makedirs(data_path / OperationStorage.UPLOAD_PATH, exist_ok=True)

    @property
    def data_path(self) -> Path:
        return self._data_path

    @property
    def upload_path(self) -> Path:
        return self.data_path / "upload_path"

    def cleanup(self) -> None:
        shutil.rmtree(self.data_path, ignore_errors=True)

        parent = self.data_path.parent
        files = os.listdir(parent)

        if len(files) == 0:
            try:
                os.rmdir(parent)
            except OSError:
                logger.debug(f"Cannot remove directory: {parent}")



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["send_checkpoint_created"]

import logging

_logger = logging.getLogger(__name__)


class MessageType(object):
    CHECKPOINT_CREATED = "CHECKPOINT_CREATED"


def send_checkpoint_created(notebook_id, notebook_path, checkpoint_id) -> None:
    """Send checkpoint created message.

    Args:
        notebook_id (:obj:`str`): The notebook's id.
        notebook_path (:obj:`str`): The notebook's path.
        checkpoint_id (:obj:`str`): The checkpoint's path.


    Raises:
        `ImportError`: If ipykernel is not available.
    """
    neptune_comm = _get_comm()
    neptune_comm.send(
        data=dict(
            message_type=MessageType.CHECKPOINT_CREATED,
            data=dict(
                checkpoint_id=str(checkpoint_id),
                notebook_id=str(notebook_id),
                notebook_path=str(notebook_path),
            ),
        )
    )


def _get_comm():
    from ipykernel.comm import Comm

    return Comm(target_name="neptune_comm")



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["with_api_exceptions_handler"]

import itertools
import logging
import os
import time

import requests
from bravado.exception import (
    BravadoConnectionError,
    BravadoTimeoutError,
    HTTPBadGateway,
    HTTPClientError,
    HTTPForbidden,
    HTTPGatewayTimeout,
    HTTPInternalServerError,
    HTTPRequestTimeout,
    HTTPServiceUnavailable,
    HTTPTooManyRequests,
    HTTPUnauthorized,
)
from requests.exceptions import ChunkedEncodingError
from urllib3.exceptions import NewConnectionError

from neptune.common.envs import NEPTUNE_RETRIES_TIMEOUT_ENV
from neptune.common.exceptions import (
    ClientHttpError,
    Forbidden,
    NeptuneAuthTokenExpired,
    NeptuneConnectionLostException,
    NeptuneInvalidApiTokenException,
    NeptuneSSLVerificationError,
    Unauthorized,
)
from neptune.common.utils import reset_internal_ssl_state

_logger = logging.getLogger(__name__)

MAX_RETRY_TIME = 30
retries_timeout = int(os.getenv(NEPTUNE_RETRIES_TIMEOUT_ENV, "60"))


def with_api_exceptions_handler(func):
    def wrapper(*args, **kwargs):
        ssl_error_occurred = False
        last_exception = None
        start_time = time.monotonic()
        for retry in itertools.count(0):
            if time.monotonic() - start_time > retries_timeout:
                break

            try:
                return func(*args, **kwargs)
            except requests.exceptions.InvalidHeader as e:
                if "X-Neptune-Api-Token" in e.args[0]:
                    raise NeptuneInvalidApiTokenException()
                raise
            except requests.exceptions.SSLError as e:
                """
                OpenSSL's internal random number generator does not properly handle forked processes.
                Applications must change the PRNG state of the parent process
                if they use any SSL feature with os.fork().
                Any successful call of RAND_add(), RAND_bytes() or RAND_pseudo_bytes() is sufficient.
                https://docs.python.org/3/library/ssl.html#multi-processing

                On Linux it looks like it does not help much but does not break anything either.
                But single retry seems to solve the issue.
                """
                if not ssl_error_occurred:
                    ssl_error_occurred = True
                    reset_internal_ssl_state()
                    continue
                raise NeptuneSSLVerificationError() from e
            except (
                BravadoConnectionError,
                BravadoTimeoutError,
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
                HTTPRequestTimeout,
                HTTPServiceUnavailable,
                HTTPGatewayTimeout,
                HTTPBadGateway,
                HTTPTooManyRequests,
                HTTPInternalServerError,
                NewConnectionError,
                ChunkedEncodingError,
            ) as e:
                time.sleep(min(2 ** min(10, retry), MAX_RETRY_TIME))
                last_exception = e
                continue
            except NeptuneAuthTokenExpired:
                continue
            except HTTPUnauthorized:
                raise Unauthorized()
            except HTTPForbidden:
                raise Forbidden()
            except HTTPClientError as e:
                raise ClientHttpError(e.status_code, e.response.text) from e
            except requests.exceptions.RequestException as e:
                if e.response is None:
                    raise
                status_code = e.response.status_code
                if status_code in (
                    HTTPRequestTimeout.status_code,
                    HTTPBadGateway.status_code,
                    HTTPServiceUnavailable.status_code,
                    HTTPGatewayTimeout.status_code,
                    HTTPTooManyRequests.status_code,
                    HTTPInternalServerError.status_code,
                ):
                    time.sleep(min(2 ** min(10, retry), MAX_RETRY_TIME))
                    last_exception = e
                    continue
                elif status_code == HTTPUnauthorized.status_code:
                    raise Unauthorized()
                elif status_code == HTTPForbidden.status_code:
                    raise Forbidden()
                elif 400 <= status_code < 500:
                    raise ClientHttpError(status_code, e.response.text) from e
                else:
                    raise
        raise NeptuneConnectionLostException(last_exception) from last_exception

    return wrapper



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    "get_metadata_container",
    "get_project",
    "get_qualified_name",
    "is_container_synced_and_remove_junk",
    "get_offline_dirs",
    "iterate_containers",
    "split_dir_name",
]

import logging
import os
import textwrap
import threading
from pathlib import Path
from typing import (
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

from neptune.common.exceptions import NeptuneException
from neptune.constants import OFFLINE_DIRECTORY
from neptune.envs import PROJECT_ENV_NAME
from neptune.exceptions import (
    MetadataContainerNotFound,
    ProjectNotFound,
)
from neptune.internal.backends.api_model import (
    ApiExperiment,
    Project,
)
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.container_type import ContainerType
from neptune.internal.disk_queue import DiskQueue
from neptune.internal.id_formats import (
    QualifiedName,
    UniqueId,
)
from neptune.internal.operation import Operation
from neptune.internal.utils.logger import logger


def get_metadata_container(
    backend: NeptuneBackend,
    container_id: Union[UniqueId, QualifiedName],
    container_type: Optional[ContainerType] = None,
) -> Optional[ApiExperiment]:
    public_container_type = container_type or "object"
    try:
        return backend.get_metadata_container(container_id, expected_container_type=container_type)
    except MetadataContainerNotFound:
        logger.warning("Can't fetch %s %s. Skipping.", public_container_type, container_id)
    except NeptuneException as e:
        logger.warning("Exception while fetching %s %s. Skipping.", public_container_type, container_id)
        logging.exception(e)

    return None


_project_name_missing_message = (
    "Project name not provided. Could not synchronize offline runs."
    " To synchronize an offline run, specify the project name with the --project flag"
    f" or by setting the {PROJECT_ENV_NAME} environment variable."
)


def _project_not_found_message(project_name: QualifiedName) -> str:
    return (
        f"Project {project_name} not found. Could not synchronize offline runs."
        " Please ensure you specified the correct project name with the --project flag"
        f" or with the {PROJECT_ENV_NAME} environment variable, or contact Neptune for support."
    )


def get_project(project_name_flag: Optional[QualifiedName], backend: NeptuneBackend) -> Optional[Project]:
    project_name = project_name_flag or QualifiedName(os.getenv(PROJECT_ENV_NAME))
    if not project_name:
        logger.warning(textwrap.fill(_project_name_missing_message))
        return None
    try:
        return backend.get_project(project_name)
    except ProjectNotFound:
        logger.warning(textwrap.fill(_project_not_found_message(project_name)))
        return None


def get_qualified_name(experiment: ApiExperiment) -> QualifiedName:
    return QualifiedName(f"{experiment.workspace}/{experiment.project_name}/{experiment.sys_id}")


def is_container_synced_and_remove_junk(experiment_path: Path) -> bool:
    return all(_is_execution_synced_and_remove_junk(execution_path) for execution_path in experiment_path.iterdir())


def _is_execution_synced_and_remove_junk(execution_path: Path) -> bool:
    """
    The DiskQueue.close() method removes junk metadata from the disk when the queue is empty.
    """
    with DiskQueue(execution_path, lambda x: x.to_dict(), Operation.from_dict, threading.RLock()) as disk_queue:
        return disk_queue.is_empty()


def split_dir_name(dir_name: str) -> Tuple[ContainerType, UniqueId]:
    parts = dir_name.split("__")
    if len(parts) == 2:
        return ContainerType(parts[0]), UniqueId(parts[1])
    elif len(parts) == 1:
        return ContainerType.RUN, UniqueId(dir_name)
    else:
        raise ValueError(f"Wrong dir format: {dir_name}")


def iterate_containers(
    base_path: Path,
) -> Iterator[Tuple[ContainerType, UniqueId, Path]]:
    if not base_path.is_dir():
        return

    for path in base_path.iterdir():
        container_type, unique_id = split_dir_name(dir_name=path.name)

        yield container_type, unique_id, path


def get_offline_dirs(base_path: Path) -> List[UniqueId]:
    result = []
    if not (base_path / OFFLINE_DIRECTORY).is_dir():
        return []
    for path_ in (base_path / OFFLINE_DIRECTORY).iterdir():
        dir_name = path_.name
        result.append(UniqueId(dir_name))
    return result



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import functools
import glob as globlib
import logging
import math
import os
import re
import ssl
import sys

import numpy as np
import pandas as pd

from neptune.common import envs
from neptune.common.exceptions import (
    FileNotFound,
    InvalidNotebookPath,
    NeptuneIncorrectProjectQualifiedNameException,
    NeptuneMissingProjectQualifiedNameException,
    NotADirectory,
    NotAFile,
)
from neptune.common.git_info import GitInfo
from neptune.common.patterns import PROJECT_QUALIFIED_NAME_PATTERN

_logger = logging.getLogger(__name__)

IS_WINDOWS = sys.platform == "win32"
IS_MACOS = sys.platform == "darwin"


def reset_internal_ssl_state():
    """
    OpenSSL's internal random number generator does not properly handle forked processes.
    Applications must change the PRNG state of the parent process if they use any SSL feature with os.fork().
    Any successful call of RAND_add(), RAND_bytes() or RAND_pseudo_bytes() is sufficient.
    https://docs.python.org/3/library/ssl.html#multi-processing
    """
    ssl.RAND_bytes(100)


def map_values(f_value, dictionary):
    return dict((k, f_value(v)) for k, v in dictionary.items())


def map_keys(f_key, dictionary):
    return dict((f_key(k), v) for k, v in dictionary.items())


def as_list(value):
    if value is None or isinstance(value, list):
        return value
    else:
        return [value]


def validate_notebook_path(path):
    if not path.endswith(".ipynb"):
        raise InvalidNotebookPath(path)

    if not os.path.exists(path):
        raise FileNotFound(path)

    if not os.path.isfile(path):
        raise NotAFile(path)


def assure_directory_exists(destination_dir):
    """Checks if `destination_dir` DIRECTORY exists, or creates one"""
    if not destination_dir:
        destination_dir = os.getcwd()

    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)
    elif not os.path.isdir(destination_dir):
        raise NotADirectory(destination_dir)

    return destination_dir


def align_channels_on_x(dataframe):
    channel_dfs, common_x = _split_df_by_stems(dataframe)
    return merge_dataframes([common_x] + channel_dfs, on="x", how="outer")


def get_channel_name_stems(columns):
    return list(set([col[2:] for col in columns]))


def merge_dataframes(dataframes, on, how="outer"):
    merged_df = functools.reduce(lambda left, right: pd.merge(left, right, on=on, how=how), dataframes)
    return merged_df


def is_float(value):
    try:
        _ = float(value)
    except ValueError:
        return False
    else:
        return True


def is_nan_or_inf(value):
    return math.isnan(value) or math.isinf(value)


def is_notebook():
    try:
        get_ipython  # noqa: F821
        return True
    except Exception:
        return False


def _split_df_by_stems(df):
    channel_dfs, x_vals = [], []
    for stem in get_channel_name_stems(df.columns):
        channel_df = df[["x_{}".format(stem), "y_{}".format(stem)]]
        channel_df.columns = ["x", stem]
        channel_df = channel_df.dropna()
        channel_dfs.append(channel_df)
        x_vals.extend(channel_df["x"].tolist())
    common_x = pd.DataFrame({"x": np.unique(x_vals)}, dtype=float)
    return channel_dfs, common_x


def discover_git_repo_location():
    import __main__

    if hasattr(__main__, "__file__"):
        return os.path.dirname(os.path.abspath(__main__.__file__))
    return None


def update_session_proxies(session, proxies):
    if proxies is not None:
        try:
            session.proxies.update(proxies)
        except (TypeError, ValueError):
            raise ValueError("Wrong proxies format: {}".format(proxies))


def get_git_info(repo_path=None):
    """Retrieve information about git repository.

    If the attempt fails, ``None`` will be returned.

    Args:
        repo_path (:obj:`str`, optional, default is ``None``):

            | Path to the repository from which extract information about git.
            | If ``None`` is passed, calling ``get_git_info`` is equivalent to calling
              ``git.Repo(search_parent_directories=True)``.
              Check `GitPython <https://gitpython.readthedocs.io/en/stable/reference.html#git.repo.base.Repo>`_
              docs for more information.

    Returns:
        :class:`~neptune.git_info.GitInfo` - An object representing information about git repository.

    Examples:

        .. code:: python3

            # Get git info from the current directory
            git_info = get_git_info('.')

    """
    try:
        import git

        repo = git.Repo(repo_path, search_parent_directories=True)

        commit = repo.head.commit

        active_branch = ""

        try:
            active_branch = repo.active_branch.name
        except TypeError as e:
            if str(e.args[0]).startswith("HEAD is a detached symbolic reference as it points to"):
                active_branch = "Detached HEAD"

        remote_urls = [remote.url for remote in repo.remotes]

        return GitInfo(
            commit_id=commit.hexsha,
            message=commit.message,
            author_name=commit.author.name,
            author_email=commit.author.email,
            commit_date=commit.committed_datetime,
            repository_dirty=repo.is_dirty(index=False, untracked_files=True),
            active_branch=active_branch,
            remote_urls=remote_urls,
        )
    except:  # noqa: E722
        return None


def file_contains(filename, text):
    with open(filename) as f:
        for line in f:
            if text in line:
                return True
        return False


def in_docker():
    cgroup_file = "/proc/self/cgroup"
    return os.path.exists("./dockerenv") or (os.path.exists(cgroup_file) and file_contains(cgroup_file, text="docker"))


def is_ipython():
    try:
        import IPython

        ipython = IPython.core.getipython.get_ipython()
        return ipython is not None
    except ImportError:
        return False


def glob(pathname):
    if sys.version_info.major < 3 or (sys.version_info.major == 3 and sys.version_info.minor < 5):
        return globlib.glob(pathname)
    else:
        return globlib.glob(pathname, recursive=True)


def assure_project_qualified_name(project_qualified_name):
    project_qualified_name = project_qualified_name or os.getenv(envs.PROJECT_ENV_NAME)

    if not project_qualified_name:
        raise NeptuneMissingProjectQualifiedNameException()
    if not re.match(PROJECT_QUALIFIED_NAME_PATTERN, project_qualified_name):
        raise NeptuneIncorrectProjectQualifiedNameException(project_qualified_name)

    return project_qualified_name


class NoopObject(object):
    def __getattr__(self, name):
        return self

    def __getitem__(self, key):
        return self

    def __call__(self, *args, **kwargs):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["NeptuneBackendMock"]

import os
import uuid
from collections import defaultdict
from datetime import datetime
from shutil import copyfile
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from zipfile import ZipFile

from neptune.api.dtos import FileEntry
from neptune.common.exceptions import (
    InternalClientError,
    NeptuneException,
)
from neptune.exceptions import (
    ContainerUUIDNotFound,
    MetadataInconsistency,
    ModelVersionNotFound,
    ProjectNotFound,
    RunNotFound,
)
from neptune.internal.artifacts.types import ArtifactFileData
from neptune.internal.backends.api_model import (
    ApiExperiment,
    ArtifactAttribute,
    Attribute,
    AttributeType,
    BoolAttribute,
    DatetimeAttribute,
    FileAttribute,
    FloatAttribute,
    FloatPointValue,
    FloatSeriesAttribute,
    FloatSeriesValues,
    ImageSeriesValues,
    IntAttribute,
    LeaderboardEntry,
    Project,
    StringAttribute,
    StringPointValue,
    StringSeriesAttribute,
    StringSeriesValues,
    StringSetAttribute,
    Workspace,
)
from neptune.internal.backends.hosted_file_operations import get_unique_upload_entries
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.backends.nql import NQLQuery
from neptune.internal.container_structure import ContainerStructure
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import (
    QualifiedName,
    SysId,
    UniqueId,
)
from neptune.internal.operation import (
    AddStrings,
    AssignArtifact,
    AssignBool,
    AssignDatetime,
    AssignFloat,
    AssignInt,
    AssignString,
    ClearArtifact,
    ClearFloatLog,
    ClearImageLog,
    ClearStringLog,
    ClearStringSet,
    ConfigFloatSeries,
    CopyAttribute,
    DeleteAttribute,
    DeleteFiles,
    LogFloats,
    LogImages,
    LogStrings,
    Operation,
    RemoveStrings,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent,
    UploadFileSet,
)
from neptune.internal.operation_processors.operation_storage import OperationStorage
from neptune.internal.operation_visitor import OperationVisitor
from neptune.internal.types.file_types import FileType
from neptune.internal.utils import base64_decode
from neptune.internal.utils.generic_attribute_mapper import NoValue
from neptune.internal.utils.git import GitInfo
from neptune.internal.utils.paths import path_to_str
from neptune.types import (
    Boolean,
    Integer,
)
from neptune.types.atoms import GitRef
from neptune.types.atoms.artifact import Artifact
from neptune.types.atoms.datetime import Datetime
from neptune.types.atoms.file import File
from neptune.types.atoms.float import Float
from neptune.types.atoms.string import String
from neptune.types.file_set import FileSet
from neptune.types.namespace import Namespace
from neptune.types.series.file_series import FileSeries
from neptune.types.series.float_series import FloatSeries
from neptune.types.series.string_series import StringSeries
from neptune.types.sets.string_set import StringSet
from neptune.types.value import Value
from neptune.types.value_visitor import ValueVisitor

Val = TypeVar("Val", bound=Value)


class NeptuneBackendMock(NeptuneBackend):
    WORKSPACE_NAME = "mock-workspace"
    PROJECT_NAME = "project-placeholder"
    PROJECT_KEY = SysId("OFFLINE")
    MODEL_SYS_ID = SysId("OFFLINE-MOD")

    def __init__(self, credentials=None, proxies=None):
        self._project_id: UniqueId = UniqueId(str(uuid.uuid4()))
        self._containers: Dict[(UniqueId, ContainerType), ContainerStructure[Value, dict]] = dict()
        self._next_run = 1  # counter for runs
        self._next_model_version = defaultdict(lambda: 1)  # counter for model versions
        self._artifacts: Dict[Tuple[str, str], List[ArtifactFileData]] = dict()
        self._attribute_type_converter_value_visitor = self.AttributeTypeConverterValueVisitor()
        self._create_container(self._project_id, ContainerType.PROJECT, self.PROJECT_KEY)

    def get_display_address(self) -> str:
        return "OFFLINE"

    def get_available_projects(
        self, workspace_id: Optional[str] = None, search_term: Optional[str] = None
    ) -> List[Project]:
        return [
            Project(
                id=UniqueId(str(uuid.uuid4())),
                name=self.PROJECT_NAME,
                workspace=self.WORKSPACE_NAME,
                sys_id=self.PROJECT_KEY,
            )
        ]

    def get_available_workspaces(self) -> List[Workspace]:
        return [Workspace(id=UniqueId(str(uuid.uuid4())), name=self.WORKSPACE_NAME)]

    def _create_container(self, container_id: UniqueId, container_type: ContainerType, sys_id: SysId):
        container = self._containers.setdefault((container_id, container_type), ContainerStructure[Value, dict]())
        container.set(["sys", "id"], String(str(sys_id)))
        container.set(["sys", "state"], String("Active"))
        container.set(["sys", "owner"], String("offline_user"))
        container.set(["sys", "size"], Float(0))
        container.set(["sys", "tags"], StringSet(set()))
        container.set(["sys", "creation_time"], Datetime(datetime.now()))
        container.set(["sys", "modification_time"], Datetime(datetime.now()))
        container.set(["sys", "failed"], Boolean(False))
        if container_type == ContainerType.MODEL_VERSION:
            container.set(["sys", "model_id"], String(str(self.MODEL_SYS_ID)))
            container.set(["sys", "stage"], String("none"))
        return container

    def _get_container(self, container_id: UniqueId, container_type: ContainerType):
        key = (container_id, container_type)
        if key not in self._containers:
            raise ContainerUUIDNotFound(container_id, container_type)
        container = self._containers[(container_id, container_type)]
        return container

    def create_run(
        self,
        project_id: UniqueId,
        git_info: Optional[GitInfo] = None,
        custom_run_id: Optional[str] = None,
        notebook_id: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ) -> ApiExperiment:
        sys_id = SysId(f"{self.PROJECT_KEY}-{self._next_run}")
        self._next_run += 1
        new_run_id = UniqueId(str(uuid.uuid4()))
        self._create_container(new_run_id, ContainerType.RUN, sys_id=sys_id)
        return ApiExperiment(
            id=new_run_id,
            type=ContainerType.RUN,
            sys_id=sys_id,
            workspace=self.WORKSPACE_NAME,
            project_name=self.PROJECT_NAME,
            trashed=False,
        )

    def create_model(self, project_id: str, key: str) -> ApiExperiment:
        sys_id = SysId(f"{self.PROJECT_KEY}-{key}")
        new_run_id = UniqueId(str(uuid.uuid4()))
        self._create_container(new_run_id, ContainerType.MODEL, sys_id=sys_id)
        return ApiExperiment(
            id=new_run_id,
            type=ContainerType.MODEL,
            sys_id=sys_id,
            workspace=self.WORKSPACE_NAME,
            project_name=self.PROJECT_NAME,
            trashed=False,
        )

    def create_model_version(self, project_id: str, model_id: UniqueId) -> ApiExperiment:
        try:
            model_key = self._get_container(container_id=model_id, container_type=ContainerType.MODEL).get("sys/id")
        except ContainerUUIDNotFound:
            model_key = "MOD"

        sys_id = SysId(f"{self.PROJECT_KEY}-{model_key}-{self._next_model_version[model_id]}")
        self._next_model_version[model_id] += 1
        new_run_id = UniqueId(str(uuid.uuid4()))
        self._create_container(new_run_id, ContainerType.MODEL_VERSION, sys_id=sys_id)
        return ApiExperiment(
            id=new_run_id,
            type=ContainerType.MODEL,
            sys_id=sys_id,
            workspace=self.WORKSPACE_NAME,
            project_name=self.PROJECT_NAME,
            trashed=False,
        )

    def create_checkpoint(self, notebook_id: str, jupyter_path: str) -> Optional[str]:
        return None

    def get_project(self, project_id: QualifiedName) -> Project:
        return Project(
            id=self._project_id,
            name=self.PROJECT_NAME,
            workspace=self.WORKSPACE_NAME,
            sys_id=self.PROJECT_KEY,
        )

    def get_metadata_container(
        self,
        container_id: Union[UniqueId, QualifiedName],
        expected_container_type: ContainerType,
    ) -> ApiExperiment:
        if "/" not in container_id:
            raise ValueError("Backend mock expect container_id as QualifiedName only")

        if expected_container_type == ContainerType.RUN:
            raise RunNotFound(container_id)
        elif expected_container_type == ContainerType.MODEL:
            return ApiExperiment(
                id=UniqueId(str(uuid.uuid4())),
                type=ContainerType.MODEL,
                sys_id=SysId(container_id.rsplit("/", 1)[-1]),
                workspace=self.WORKSPACE_NAME,
                project_name=self.PROJECT_NAME,
            )
        elif expected_container_type == ContainerType.MODEL_VERSION:
            raise ModelVersionNotFound(container_id)
        else:
            raise ProjectNotFound(container_id)

    def execute_operations(
        self,
        container_id: UniqueId,
        container_type: ContainerType,
        operations: List[Operation],
        operation_storage: OperationStorage,
    ) -> Tuple[int, List[NeptuneException]]:
        result = []
        for op in operations:
            try:
                self._execute_operation(container_id, container_type, op, operation_storage)
            except NeptuneException as e:
                result.append(e)
        return len(operations), result

    def _execute_operation(
        self, container_id: UniqueId, container_type: ContainerType, op: Operation, operation_storage: OperationStorage
    ) -> None:
        run = self._get_container(container_id, container_type)
        val = run.get(op.path)
        if val is not None and not isinstance(val, Value):
            if isinstance(val, dict):
                raise MetadataInconsistency("{} is a namespace, not an attribute".format(op.path))
            else:
                raise InternalClientError("{} is a {}".format(op.path, type(val)))
        visitor = NeptuneBackendMock.NewValueOpVisitor(self, op.path, val, operation_storage)
        new_val = visitor.visit(op)
        if new_val is not None:
            run.set(op.path, new_val)
        else:
            run.pop(op.path)

    def get_attributes(self, container_id: str, container_type: ContainerType) -> List[Attribute]:
        run = self._get_container(container_id, container_type)
        return list(self._generate_attributes(None, run.get_structure()))

    def _generate_attributes(self, base_path: Optional[str], values: dict):
        for key, value_or_dict in values.items():
            new_path = base_path + "/" + key if base_path is not None else key
            if isinstance(value_or_dict, dict):
                yield from self._generate_attributes(new_path, value_or_dict)
            else:
                yield Attribute(
                    new_path,
                    value_or_dict.accept(self._attribute_type_converter_value_visitor),
                )

    def download_file(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        destination: Optional[str] = None,
    ):
        run = self._get_container(container_id, container_type)
        value: File = run.get(path)
        target_path = os.path.abspath(destination or (path[-1] + ("." + value.extension if value.extension else "")))
        if value.file_type is FileType.IN_MEMORY:
            with open(target_path, "wb") as target_file:
                target_file.write(value.content)
        elif value.file_type is FileType.LOCAL_FILE:
            if value.path != target_path:
                copyfile(value.path, target_path)
        else:
            raise ValueError(f"Unexpected FileType: {value.file_type}")

    def download_file_set(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        destination: Optional[str] = None,
    ):
        run = self._get_container(container_id, container_type)
        source_file_set_value: FileSet = run.get(path)

        if destination is None:
            target_file = path[-1] + ".zip"
        elif os.path.isdir(destination):
            target_file = os.path.join(destination, path[-1] + ".zip")
        else:
            target_file = destination

        upload_entries = get_unique_upload_entries(source_file_set_value.file_globs)

        with ZipFile(target_file, "w") as zipObj:
            for upload_entry in upload_entries:
                zipObj.write(upload_entry.source, upload_entry.target_path)

    def get_float_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> FloatAttribute:
        val = self._get_attribute(container_id, container_type, path, Float)
        return FloatAttribute(val.value)

    def get_int_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> IntAttribute:
        val = self._get_attribute(container_id, container_type, path, Integer)
        return IntAttribute(val.value)

    def get_bool_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> BoolAttribute:
        val = self._get_attribute(container_id, container_type, path, Boolean)
        return BoolAttribute(val.value)

    def get_file_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> FileAttribute:
        val = self._get_attribute(container_id, container_type, path, File)
        return FileAttribute(
            name=os.path.basename(val.path) if val.file_type is FileType.LOCAL_FILE else "",
            ext=val.extension or "",
            size=0,
        )

    def get_string_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringAttribute:
        val = self._get_attribute(container_id, container_type, path, String)
        return StringAttribute(val.value)

    def get_datetime_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> DatetimeAttribute:
        val = self._get_attribute(container_id, container_type, path, Datetime)
        return DatetimeAttribute(val.value)

    def get_artifact_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> ArtifactAttribute:
        val = self._get_attribute(container_id, container_type, path, Artifact)
        return ArtifactAttribute(val.hash)

    def list_artifact_files(self, project_id: str, artifact_hash: str) -> List[ArtifactFileData]:
        return self._artifacts[(project_id, artifact_hash)]

    def get_float_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> FloatSeriesAttribute:
        val = self._get_attribute(container_id, container_type, path, FloatSeries)
        return FloatSeriesAttribute(val.values[-1] if val.values else None)

    def get_string_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSeriesAttribute:
        val = self._get_attribute(container_id, container_type, path, StringSeries)
        return StringSeriesAttribute(val.values[-1] if val.values else None)

    def get_string_set_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSetAttribute:
        val = self._get_attribute(container_id, container_type, path, StringSet)
        return StringSetAttribute(set(val.values))

    def _get_attribute(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        expected_type: Type[Val],
    ) -> Val:
        run = self._get_container(container_id, container_type)
        value: Optional[Value] = run.get(path)
        str_path = path_to_str(path)
        if value is None:
            raise MetadataInconsistency("Attribute {} not found".format(str_path))
        if isinstance(value, expected_type):
            return value
        raise MetadataInconsistency("Attribute {} is not {}".format(str_path, type.__name__))

    def get_string_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        offset: int,
        limit: int,
    ) -> StringSeriesValues:
        val = self._get_attribute(container_id, container_type, path, StringSeries)
        return StringSeriesValues(
            len(val.values),
            [StringPointValue(timestampMillis=42342, step=idx, value=v) for idx, v in enumerate(val.values)],
        )

    def get_float_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        offset: int,
        limit: int,
    ) -> FloatSeriesValues:
        val = self._get_attribute(container_id, container_type, path, FloatSeries)
        return FloatSeriesValues(
            len(val.values),
            [FloatPointValue(timestampMillis=42342, step=idx, value=v) for idx, v in enumerate(val.values)],
        )

    def get_image_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        offset: int,
        limit: int,
    ) -> ImageSeriesValues:
        return ImageSeriesValues(0)

    def download_file_series_by_index(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        index: int,
        destination: str,
    ):
        """Non relevant for backend"""

    def get_run_url(self, run_id: str, workspace: str, project_name: str, sys_id: str) -> str:
        return f"offline/{run_id}"

    def get_project_url(self, project_id: str, workspace: str, project_name: str) -> str:
        return f"offline/{project_id}"

    def get_model_url(self, model_id: str, workspace: str, project_name: str, sys_id: str) -> str:
        return f"offline/{model_id}"

    def get_model_version_url(
        self,
        model_version_id: str,
        model_id: str,
        workspace: str,
        project_name: str,
        sys_id: str,
    ) -> str:
        return f"offline/{model_version_id}"

    def _get_attribute_values(self, value_dict, path_prefix: List[str]):
        assert isinstance(value_dict, dict)
        for k, value in value_dict.items():
            if isinstance(value, dict):
                yield from self._get_attribute_values(value, path_prefix + [k])
            else:
                attr_type = value.accept(self._attribute_type_converter_value_visitor).value
                attr_path = "/".join(path_prefix + [k])
                if hasattr(value, "value"):
                    yield attr_path, attr_type, value.value
                else:
                    return attr_path, attr_type, NoValue

    def fetch_atom_attribute_values(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> List[Tuple[str, AttributeType, Any]]:
        run = self._get_container(container_id, container_type)
        values = self._get_attribute_values(run.get(path), path)
        namespace_prefix = path_to_str(path)
        if namespace_prefix:
            # don't want to catch "ns/attribute/other" while looking for "ns/attr"
            namespace_prefix += "/"
        return [
            (full_path, attr_type, attr_value)
            for (full_path, attr_type, attr_value) in values
            if full_path.startswith(namespace_prefix)
        ]

    def search_leaderboard_entries(
        self,
        project_id: UniqueId,
        types: Optional[Iterable[ContainerType]] = None,
        query: Optional[NQLQuery] = None,
        columns: Optional[Iterable[str]] = None,
    ) -> List[LeaderboardEntry]:
        """Non relevant for mock"""

    class AttributeTypeConverterValueVisitor(ValueVisitor[AttributeType]):
        def visit_float(self, _: Float) -> AttributeType:
            return AttributeType.FLOAT

        def visit_integer(self, _: Integer) -> AttributeType:
            return AttributeType.INT

        def visit_boolean(self, _: Boolean) -> AttributeType:
            return AttributeType.BOOL

        def visit_string(self, _: String) -> AttributeType:
            return AttributeType.STRING

        def visit_datetime(self, _: Datetime) -> AttributeType:
            return AttributeType.DATETIME

        def visit_file(self, _: File) -> AttributeType:
            return AttributeType.FILE

        def visit_file_set(self, _: FileSet) -> AttributeType:
            return AttributeType.FILE_SET

        def visit_float_series(self, _: FloatSeries) -> AttributeType:
            return AttributeType.FLOAT_SERIES

        def visit_string_series(self, _: StringSeries) -> AttributeType:
            return AttributeType.STRING_SERIES

        def visit_image_series(self, _: FileSeries) -> AttributeType:
            return AttributeType.IMAGE_SERIES

        def visit_string_set(self, _: StringSet) -> AttributeType:
            return AttributeType.STRING_SET

        def visit_git_ref(self, _: GitRef) -> AttributeType:
            return AttributeType.GIT_REF

        def visit_artifact(self, _: Artifact) -> AttributeType:
            return AttributeType.ARTIFACT

        def visit_namespace(self, _: Namespace) -> AttributeType:
            raise NotImplementedError

        def copy_value(self, source_type: Type[Attribute], source_path: List[str]) -> AttributeType:
            raise NotImplementedError

    class NewValueOpVisitor(OperationVisitor[Optional[Value]]):
        def __init__(
            self, backend, path: List[str], current_value: Optional[Value], operation_storage: OperationStorage
        ):
            self._backend = backend
            self._path = path
            self._current_value = current_value
            self._artifact_hash = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
            self._operation_storage = operation_storage

        def visit_assign_float(self, op: AssignFloat) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Float):
                raise self._create_type_error("assign", Float.__name__)
            return Float(op.value)

        def visit_assign_int(self, op: AssignInt) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Integer):
                raise self._create_type_error("assign", Integer.__name__)
            return Integer(op.value)

        def visit_assign_bool(self, op: AssignBool) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Boolean):
                raise self._create_type_error("assign", Boolean.__name__)
            return Boolean(op.value)

        def visit_assign_string(self, op: AssignString) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, String):
                raise self._create_type_error("assign", String.__name__)
            return String(op.value)

        def visit_assign_datetime(self, op: AssignDatetime) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Datetime):
                raise self._create_type_error("assign", Datetime.__name__)
            return Datetime(op.value)

        def visit_assign_artifact(self, op: AssignArtifact) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Artifact):
                raise self._create_type_error("assign", Artifact.__name__)
            return Artifact(op.hash)

        def visit_track_files_to_artifact(self, _: TrackFilesToArtifact) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Artifact):
                raise self._create_type_error("save", Artifact.__name__)
            return Artifact(self._artifact_hash)

        def visit_clear_artifact(self, _: ClearArtifact) -> Optional[Value]:
            if self._current_value is None:
                return Artifact()
            if not isinstance(self._current_value, Artifact):
                raise self._create_type_error("clear", Artifact.__name__)
            return Artifact()

        def visit_upload_file(self, op: UploadFile) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, File):
                raise self._create_type_error("save", File.__name__)
            return File.from_path(path=op.get_absolute_path(self._operation_storage), extension=op.ext)

        def visit_upload_file_content(self, op: UploadFileContent) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, File):
                raise self._create_type_error("upload_files", File.__name__)
            return File.from_content(content=base64_decode(op.file_content), extension=op.ext)

        def visit_upload_file_set(self, op: UploadFileSet) -> Optional[Value]:
            if self._current_value is None or op.reset:
                return FileSet(op.file_globs)
            if not isinstance(self._current_value, FileSet):
                raise self._create_type_error("save", FileSet.__name__)
            return FileSet(self._current_value.file_globs + op.file_globs)

        def visit_log_floats(self, op: LogFloats) -> Optional[Value]:
            raw_values = [x.value for x in op.values]
            if self._current_value is None:
                return FloatSeries(raw_values)
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("log", FloatSeries.__name__)
            return FloatSeries(
                self._current_value.values + raw_values,
                min=self._current_value.min,
                max=self._current_value.max,
                unit=self._current_value.unit,
            )

        def visit_log_strings(self, op: LogStrings) -> Optional[Value]:
            raw_values = [x.value for x in op.values]
            if self._current_value is None:
                return StringSeries(raw_values)
            if not isinstance(self._current_value, StringSeries):
                raise self._create_type_error("log", StringSeries.__name__)
            return StringSeries(self._current_value.values + raw_values)

        def visit_log_images(self, op: LogImages) -> Optional[Value]:
            raw_values = [File.from_content(base64_decode(x.value.data)) for x in op.values]
            if self._current_value is None:
                return FileSeries(raw_values)
            if not isinstance(self._current_value, FileSeries):
                raise self._create_type_error("log", FileSeries.__name__)
            return FileSeries(self._current_value.values + raw_values)

        def visit_clear_float_log(self, op: ClearFloatLog) -> Optional[Value]:
            if self._current_value is None:
                return FloatSeries([])
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("clear", FloatSeries.__name__)
            return FloatSeries(
                [],
                min=self._current_value.min,
                max=self._current_value.max,
                unit=self._current_value.unit,
            )

        def visit_clear_string_log(self, op: ClearStringLog) -> Optional[Value]:
            if self._current_value is None:
                return StringSeries([])
            if not isinstance(self._current_value, StringSeries):
                raise self._create_type_error("clear", StringSeries.__name__)
            return StringSeries([])

        def visit_clear_image_log(self, op: ClearImageLog) -> Optional[Value]:
            if self._current_value is None:
                return FileSeries([])
            if not isinstance(self._current_value, FileSeries):
                raise self._create_type_error("clear", FileSeries.__name__)
            return FileSeries([])

        def visit_config_float_series(self, op: ConfigFloatSeries) -> Optional[Value]:
            if self._current_value is None:
                return FloatSeries([], min=op.min, max=op.max, unit=op.unit)
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("log", FloatSeries.__name__)
            return FloatSeries(self._current_value.values, min=op.min, max=op.max, unit=op.unit)

        def visit_add_strings(self, op: AddStrings) -> Optional[Value]:
            if self._current_value is None:
                return StringSet(op.values)
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("add", StringSet.__name__)
            return StringSet(self._current_value.values.union(op.values))

        def visit_remove_strings(self, op: RemoveStrings) -> Optional[Value]:
            if self._current_value is None:
                return StringSet(set())
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("remove", StringSet.__name__)
            return StringSet(self._current_value.values.difference(op.values))

        def visit_clear_string_set(self, op: ClearStringSet) -> Optional[Value]:
            if self._current_value is None:
                return StringSet(set())
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("clear", StringSet.__name__)
            return StringSet(set())

        def visit_delete_files(self, op: DeleteFiles) -> Optional[Value]:
            if self._current_value is None:
                return FileSet([])
            if not isinstance(self._current_value, FileSet):
                raise self._create_type_error("delete_files", FileSet.__name__)
            # It is not important to support deleting properly in debug mode, let's just ignore this operation
            return self._current_value

        def visit_delete_attribute(self, op: DeleteAttribute) -> Optional[Value]:
            if self._current_value is None:
                raise MetadataInconsistency(
                    "Cannot perform delete operation on {}. Attribute is undefined.".format(self._path)
                )
            return None

        def visit_copy_attribute(self, op: CopyAttribute) -> Optional[Value]:
            return op.resolve(self._backend).accept(self)

        def _create_type_error(self, op_name, expected):
            return MetadataInconsistency(
                "Cannot perform {} operation on {}. Expected {}, {} found.".format(
                    op_name, self._path, expected, type(self._current_value)
                )
            )

    def list_fileset_files(self, attribute: List[str], container_id: str, path: str) -> List[FileEntry]:
        return [
            FileEntry(
                name="mock_name",
                size=100,
                mtime=datetime.now(),
                file_type="file",
            )
        ]



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Run"]

import os
import threading
from platform import node as get_hostname
from typing import (
    List,
    Optional,
    Tuple,
    Union,
)

from neptune.attributes.constants import (
    SYSTEM_DESCRIPTION_ATTRIBUTE_PATH,
    SYSTEM_FAILED_ATTRIBUTE_PATH,
    SYSTEM_HOSTNAME_ATTRIBUTE_PATH,
    SYSTEM_NAME_ATTRIBUTE_PATH,
    SYSTEM_TAGS_ATTRIBUTE_PATH,
)
from neptune.common.warnings import (
    NeptuneWarning,
    warn_once,
)
from neptune.envs import (
    CONNECTION_MODE,
    CUSTOM_RUN_ID_ENV_NAME,
    MONITORING_NAMESPACE,
    NEPTUNE_NOTEBOOK_ID,
    NEPTUNE_NOTEBOOK_PATH,
)
from neptune.exceptions import (
    InactiveRunException,
    NeedExistingRunForReadOnlyMode,
    NeptunePossibleLegacyUsageException,
    NeptuneRunResumeAndCustomIdCollision,
)
from neptune.internal.backends.api_model import ApiExperiment
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.backgroud_job_list import BackgroundJobList
from neptune.internal.container_type import ContainerType
from neptune.internal.hardware.hardware_metric_reporting_job import HardwareMetricReportingJob
from neptune.internal.id_formats import QualifiedName
from neptune.internal.init.parameters import (
    ASYNC_LAG_THRESHOLD,
    ASYNC_NO_PROGRESS_THRESHOLD,
    DEFAULT_FLUSH_PERIOD,
    DEFAULT_NAME,
    OFFLINE_PROJECT_QUALIFIED_NAME,
)
from neptune.internal.notebooks.notebooks import create_checkpoint
from neptune.internal.state import ContainerState
from neptune.internal.streams.std_capture_background_job import (
    StderrCaptureBackgroundJob,
    StdoutCaptureBackgroundJob,
)
from neptune.internal.utils import (
    verify_collection_type,
    verify_type,
)
from neptune.internal.utils.dependency_tracking import (
    FileDependenciesStrategy,
    InferDependenciesStrategy,
)
from neptune.internal.utils.git import (
    to_git_info,
    track_uncommitted_changes,
)
from neptune.internal.utils.hashing import generate_hash
from neptune.internal.utils.limits import custom_run_id_exceeds_length
from neptune.internal.utils.ping_background_job import PingBackgroundJob
from neptune.internal.utils.runningmode import (
    in_interactive,
    in_notebook,
)
from neptune.internal.utils.source_code import upload_source_code
from neptune.internal.utils.traceback_job import TracebackJob
from neptune.internal.websockets.websocket_signals_background_job import WebsocketSignalsBackgroundJob
from neptune.metadata_containers import MetadataContainer
from neptune.metadata_containers.abstract import NeptuneObjectCallback
from neptune.types import (
    GitRef,
    StringSeries,
)
from neptune.types.atoms.git_ref import GitRefDisabled
from neptune.types.mode import Mode


class Run(MetadataContainer):
    """Starts a tracked run that logs ML model-building metadata to neptune.ai."""

    container_type = ContainerType.RUN

    LEGACY_METHODS = (
        "create_experiment",
        "send_metric",
        "log_metric",
        "send_text",
        "log_text",
        "send_image",
        "log_image",
        "send_artifact",
        "log_artifact",
        "delete_artifacts",
        "download_artifact",
        "download_sources",
        "download_artifacts",
        "reset_log",
        "get_parameters",
        "get_properties",
        "set_property",
        "remove_property",
        "get_hardware_utilization",
        "get_numeric_channels_values",
    )

    def __init__(
        self,
        with_id: Optional[str] = None,
        *,
        project: Optional[str] = None,
        api_token: Optional[str] = None,
        custom_run_id: Optional[str] = None,
        mode: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Union[List[str], str]] = None,
        source_files: Optional[Union[List[str], str]] = None,
        capture_stdout: Optional[bool] = None,
        capture_stderr: Optional[bool] = None,
        capture_hardware_metrics: Optional[bool] = None,
        fail_on_exception: bool = True,
        monitoring_namespace: Optional[str] = None,
        flush_period: float = DEFAULT_FLUSH_PERIOD,
        proxies: Optional[dict] = None,
        capture_traceback: bool = True,
        git_ref: Optional[Union[GitRef, GitRefDisabled, bool]] = None,
        dependencies: Optional[Union[str, os.PathLike]] = None,
        async_lag_callback: Optional[NeptuneObjectCallback] = None,
        async_lag_threshold: float = ASYNC_LAG_THRESHOLD,
        async_no_progress_callback: Optional[NeptuneObjectCallback] = None,
        async_no_progress_threshold: float = ASYNC_NO_PROGRESS_THRESHOLD,
        **kwargs,
    ):
        """Starts a new tracked run that logs ML model-building metadata to neptune.ai.

        You can log metadata by assigning it to the initialized Run object:

        ```
        run = neptune.init_run()
        run["your/structure"] = some_metadata
        ```

        Examples of metadata you can log: metrics, losses, scores, artifact versions, images, predictions,
        model weights, parameters, checkpoints, and interactive visualizations.

        By default, the run automatically tracks hardware consumption, stdout/stderr, source code, and Git information.
        If you're using Neptune in an interactive session, however, some background monitoring needs to be enabled
        explicitly.

        If you provide the ID of an existing run, that run is resumed and no new run is created. You may resume a run
        either to log more metadata or to fetch metadata from it.

        The run ends either when its `stop()` method is called or when the script finishes execution.

        You can also use the Run object as a context manager (see examples).

        Args:
            project: Name of the project where the run should go, in the form `workspace-name/project_name`.
            If None (default), the value of the NEPTUNE_PROJECT environment variable is used.
            api_token: User's API token.
                If None (default), the value of the NEPTUNE_API_TOKEN environment variable is used.
                Note: To keep your API token secure, save it to the NEPTUNE_API_TOKEN environment variable rather than
                placing it in plain text in the source code.
            with_id: If you want to resume a run, pass the identifier of an existing run. For example, "SAN-1".
                If None (default) is passed, starts a new tracked run.
            custom_run_id: A unique identifier to be used when running Neptune in distributed training jobs.
                Make sure to use the same identifier throughout the whole pipeline execution.
            mode: Connection mode in which the tracking will work.
                If None (default), the value of the NEPTUNE_MODE environment variable is used.
                If no value was set for the environment variable, "async" is used by default.
                Possible values are `async`, `sync`, `offline`, `read-only`, and `debug`.
            name: Editable name of the run. Defaults to "Untitled".
                The name is displayed in the run details and as a column in the runs table.
            description: Editable description of the run. Defaults to `""`.
                The description is displayed in the run details and can be added to the runs table as a column.
            tags: Tags of the run as a list of strings. Defaults to `[]`.
                Tags are displayed in the run details and in the runs table as a column.
                You can edit the tags after the run is created, either through the app or the API.
            source_files: List of source files to be uploaded.
                Uploaded source files are displayed in the "Source code" tab of the run view.
                To not upload anything, pass an empty list (`[]`).
                Unix style pathname pattern expansion is supported. For example, you can pass `*.py` to upload
                all Python files from the current directory.
                If None is passed, the Python file from which the run was created will be uploaded.
            capture_stdout: Whether to log the stdout of the run.
                Defaults to `False` in interactive sessions and `True` otherwise.
                The data is logged under the monitoring namespace (see the `monitoring_namespace` parameter).
            capture_stderr: Whether to log the stderr of the run.
                Defaults to `False` in interactive sessions and `True` otherwise.
                The data is logged under the monitoring namespace (see the `monitoring_namespace` parameter).
            capture_hardware_metrics: Whether to send hardware monitoring logs (CPU, GPU, and memory utilization).
                Defaults to `False` in interactive sessions and `True` otherwise.
                The data is logged under the monitoring namespace (see the `monitoring_namespace` parameter).
            fail_on_exception: Whether to register an uncaught exception handler to this process and,
                in case of an exception, set the "sys/failed" field of the run to `True`.
                An exception is always logged.
            monitoring_namespace: Namespace inside which all hardware monitoring logs are stored.
                Defaults to "monitoring/<hash>", where the hash is generated based on environment information,
                to ensure that it's unique for each process.
            flush_period: In the asynchronous (default) connection mode, how often disk flushing is triggered
                (in seconds).
            proxies: Argument passed to HTTP calls made via the Requests library, as dictionary of strings.
                For more information about proxies, see the Requests documentation.
            capture_traceback: Whether to log the traceback of the run in case of an exception.
                The tracked metadata is stored in the "<monitoring_namespace>/traceback" namespace (see the
                `monitoring_namespace` parameter).
            git_ref: GitRef object containing information about the Git repository path.
                If None, Neptune looks for a repository in the path of the script that is executed.
                To specify a different location, set to GitRef(repository_path="path/to/repo").
                To turn off Git tracking for the run, set to False or GitRef.DISABLED.
            dependencies: If you pass `"infer"`, Neptune logs dependencies installed in the current environment.
                You can also pass a path to your dependency file directly.
                If left empty, no dependency file is uploaded.
            async_lag_callback: Custom callback which is called if the lag between a queued operation and its
                synchronization with the server exceeds the duration defined by `async_lag_threshold`. The callback
                should take a Run object as the argument and can contain any custom code, such as calling `stop()` on
                the object.
                Note: Instead of using this argument, you can use Neptune's default callback by setting the
                `NEPTUNE_ENABLE_DEFAULT_ASYNC_LAG_CALLBACK` environment variable to `TRUE`.
            async_lag_threshold: In seconds, duration between the queueing and synchronization of an operation.
                If a lag callback (default callback enabled via environment variable or custom callback passed to the
                `async_lag_callback` argument) is enabled, the callback is called when this duration is exceeded.
            async_no_progress_callback: Custom callback which is called if there has been no synchronization progress
                whatsoever for the duration defined by `async_no_progress_threshold`. The callback
                should take a Run object as the argument and can contain any custom code, such as calling `stop()` on
                the object.
                Note: Instead of using this argument, you can use Neptune's default callback by setting the
                `NEPTUNE_ENABLE_DEFAULT_ASYNC_NO_PROGRESS_CALLBACK` environment variable to `TRUE`.
            async_no_progress_threshold: In seconds, for how long there has been no synchronization progress since the
                object was initialized. If a no-progress callback (default callback enabled via environment variable or
                custom callback passed to the `async_no_progress_callback` argument) is enabled, the callback is called
                when this duration is exceeded.

        Returns:
            Run object that is used to manage the tracked run and log metadata to it.

        Examples:

            Creating a new run:

            >>> import neptune

            >>> # Minimal invoke
            ... # (creates a run in the project specified by the NEPTUNE_PROJECT environment variable)
            ... run = neptune.init_run()

            >>> # Or initialize with the constructor
            ... run = Run(project="ml-team/classification")

            >>> # Create a run with a name and description, with no sources files or Git info tracked:
            >>> run = neptune.init_run(
            ...     name="neural-net-mnist",
            ...     description="neural net trained on MNIST",
            ...     source_files=[],
            ...     git_ref=False,
            ... )

            >>> # Log all .py files from all subdirectories, excluding hidden files
            ... run = neptune.init_run(source_files="**/*.py")

            >>> # Log all files and directories in the current working directory, excluding hidden files
            ... run = neptune.init_run(source_files="*")

            >>> # Larger example
            ... run = neptune.init_run(
            ...     project="ml-team/classification",
            ...     name="first-pytorch-ever",
            ...     description="Longer description of the run goes here",
            ...     tags=["tags", "go-here", "as-list-of-strings"],
            ...     source_files=["training_with_pytorch.py", "net.py"],
            ...     dependencies="infer",
            ...     capture_stderr=False,
            ...     git_ref=GitRef(repository_path="/Users/Jackie/repos/cls_project"),
            ... )

            Connecting to an existing run:

            >>> # Resume logging to an existing run with the ID "SAN-3"
            ... run = neptune.init_run(with_id="SAN-3")
            ... run["parameters/lr"] = 0.1  # modify or add metadata

            >>> # Initialize an existing run in read-only mode (logging new data is not possible, only fetching)
            ... run = neptune.init_run(with_id="SAN-4", mode="read-only")
            ... learning_rate = run["parameters/lr"].fetch()

            Using the Run object as context manager:

            >>> with Run() as run:
            ...     run["metric"].append(value)

        For more, see the docs:
            Initializing a run:
                https://docs.neptune.ai/api/neptune#init_run
            Run class reference:
                https://docs.neptune.ai/api/run/
            Essential logging methods:
                https://docs.neptune.ai/logging/methods/
            Resuming a run:
                https://docs.neptune.ai/logging/to_existing_object/
            Setting a custom run ID:
                https://docs.neptune.ai/logging/custom_run_id/
            Logging to multiple runs at once:
                https://docs.neptune.ai/logging/to_multiple_objects/
            Accessing the run from multiple places:
                https://docs.neptune.ai/logging/from_multiple_places/
        """
        check_for_extra_kwargs("Run", kwargs)

        verify_type("with_id", with_id, (str, type(None)))
        verify_type("project", project, (str, type(None)))
        verify_type("custom_run_id", custom_run_id, (str, type(None)))
        verify_type("mode", mode, (str, type(None)))
        verify_type("name", name, (str, type(None)))
        verify_type("description", description, (str, type(None)))
        verify_type("capture_stdout", capture_stdout, (bool, type(None)))
        verify_type("capture_stderr", capture_stderr, (bool, type(None)))
        verify_type("capture_hardware_metrics", capture_hardware_metrics, (bool, type(None)))
        verify_type("fail_on_exception", fail_on_exception, bool)
        verify_type("monitoring_namespace", monitoring_namespace, (str, type(None)))
        verify_type("capture_traceback", capture_traceback, bool)
        verify_type("capture_traceback", capture_traceback, bool)
        verify_type("git_ref", git_ref, (GitRef, str, bool, type(None)))
        verify_type("dependencies", dependencies, (str, os.PathLike, type(None)))

        if tags is not None:
            if isinstance(tags, str):
                tags = [tags]
            else:
                verify_collection_type("tags", tags, str)
        if source_files is not None:
            if isinstance(source_files, str):
                source_files = [source_files]
            else:
                verify_collection_type("source_files", source_files, str)

        self._with_id: Optional[str] = with_id
        self._name: Optional[str] = DEFAULT_NAME if with_id is None and name is None else name
        self._description: Optional[str] = "" if with_id is None and description is None else description
        self._custom_run_id: Optional[str] = custom_run_id or os.getenv(CUSTOM_RUN_ID_ENV_NAME)
        self._hostname: str = get_hostname()
        self._pid: int = os.getpid()
        self._tid: int = threading.get_ident()
        self._tags: Optional[List[str]] = tags
        self._source_files: Optional[List[str]] = source_files
        self._fail_on_exception: bool = fail_on_exception
        self._capture_traceback: bool = capture_traceback

        if type(git_ref) is bool:
            if not git_ref:
                git_ref = GitRefDisabled
            else:
                git_ref = GitRef()

        self._git_ref: Optional[GitRef, GitRefDisabled] = git_ref or GitRef()
        self._dependencies: Optional[str, os.PathLike] = dependencies

        self._monitoring_namespace: str = (
            monitoring_namespace
            or os.getenv(MONITORING_NAMESPACE)
            or generate_monitoring_namespace(self._hostname, self._pid, self._tid)
        )

        # for backward compatibility imports
        mode = Mode(mode or os.getenv(CONNECTION_MODE) or Mode.ASYNC.value)

        self._stdout_path: str = "{}/stdout".format(self._monitoring_namespace)
        self._capture_stdout: bool = capture_stdout
        if capture_stdout is None:
            self._capture_stdout = capture_only_if_non_interactive(mode=mode)

        self._stderr_path: str = "{}/stderr".format(self._monitoring_namespace)
        self._capture_stderr: bool = capture_stderr
        if capture_stderr is None:
            self._capture_stderr = capture_only_if_non_interactive(mode=mode)

        self._capture_hardware_metrics: bool = capture_hardware_metrics
        if capture_hardware_metrics is None:
            self._capture_hardware_metrics = capture_only_if_non_interactive(mode=mode)

        if with_id and custom_run_id:
            raise NeptuneRunResumeAndCustomIdCollision()

        if mode == Mode.OFFLINE or mode == Mode.DEBUG:
            project = OFFLINE_PROJECT_QUALIFIED_NAME

        super().__init__(
            project=project,
            api_token=api_token,
            mode=mode,
            flush_period=flush_period,
            proxies=proxies,
            async_lag_callback=async_lag_callback,
            async_lag_threshold=async_lag_threshold,
            async_no_progress_callback=async_no_progress_callback,
            async_no_progress_threshold=async_no_progress_threshold,
        )

    def _get_or_create_api_object(self) -> ApiExperiment:
        project_workspace = self._project_api_object.workspace
        project_name = self._project_api_object.name
        project_qualified_name = f"{project_workspace}/{project_name}"

        if self._with_id:
            return self._backend.get_metadata_container(
                container_id=QualifiedName(project_qualified_name + "/" + self._with_id),
                expected_container_type=Run.container_type,
            )
        else:
            if self._mode == Mode.READ_ONLY:
                raise NeedExistingRunForReadOnlyMode()

            git_info = to_git_info(git_ref=self._git_ref)

            custom_run_id = self._custom_run_id
            if custom_run_id_exceeds_length(self._custom_run_id):
                custom_run_id = None

            notebook_id, checkpoint_id = create_notebook_checkpoint(backend=self._backend)

            return self._backend.create_run(
                project_id=self._project_api_object.id,
                git_info=git_info,
                custom_run_id=custom_run_id,
                notebook_id=notebook_id,
                checkpoint_id=checkpoint_id,
            )

    def _prepare_background_jobs(self) -> BackgroundJobList:
        background_jobs = [PingBackgroundJob()]

        websockets_factory = self._backend.websockets_factory(self._project_api_object.id, self._id)
        if websockets_factory:
            background_jobs.append(WebsocketSignalsBackgroundJob(websockets_factory))

        if self._capture_stdout:
            background_jobs.append(StdoutCaptureBackgroundJob(attribute_name=self._stdout_path))

        if self._capture_stderr:
            background_jobs.append(StderrCaptureBackgroundJob(attribute_name=self._stderr_path))

        if self._capture_hardware_metrics:
            background_jobs.append(HardwareMetricReportingJob(attribute_namespace=self._monitoring_namespace))

        if self._capture_traceback:
            background_jobs.append(
                TracebackJob(path=f"{self._monitoring_namespace}/traceback", fail_on_exception=self._fail_on_exception)
            )

        return BackgroundJobList(background_jobs)

    def _write_initial_attributes(self):
        if self._name is not None:
            self[SYSTEM_NAME_ATTRIBUTE_PATH] = self._name

        if self._description is not None:
            self[SYSTEM_DESCRIPTION_ATTRIBUTE_PATH] = self._description

        if self._hostname is not None:
            self[f"{self._monitoring_namespace}/hostname"] = self._hostname
            if self._with_id is None:
                self[SYSTEM_HOSTNAME_ATTRIBUTE_PATH] = self._hostname

        if self._pid is not None:
            self[f"{self._monitoring_namespace}/pid"] = str(self._pid)

        if self._tid is not None:
            self[f"{self._monitoring_namespace}/tid"] = str(self._tid)

        if self._tags is not None:
            self[SYSTEM_TAGS_ATTRIBUTE_PATH].add(self._tags)

        if self._with_id is None:
            self[SYSTEM_FAILED_ATTRIBUTE_PATH] = False

        if self._capture_stdout and not self.exists(self._stdout_path):
            self.define(self._stdout_path, StringSeries([]))

        if self._capture_stderr and not self.exists(self._stderr_path):
            self.define(self._stderr_path, StringSeries([]))

        if self._with_id is None or self._source_files is not None:
            # upload default sources ONLY if creating a new run
            upload_source_code(source_files=self._source_files, run=self)

        if self._dependencies:
            try:
                if self._dependencies == "infer":
                    dependency_strategy = InferDependenciesStrategy()

                else:
                    dependency_strategy = FileDependenciesStrategy(path=self._dependencies)

                dependency_strategy.log_dependencies(run=self)
            except Exception as e:
                warn_once(
                    "An exception occurred in automatic dependency tracking."
                    "Skipping upload of requirement files."
                    "Exception: " + str(e),
                    exception=NeptuneWarning,
                )

        try:
            track_uncommitted_changes(
                git_ref=self._git_ref,
                run=self,
            )
        except Exception as e:
            warn_once(
                "An exception occurred in tracking uncommitted changes."
                "Skipping upload of patch files."
                "Exception: " + str(e),
                exception=NeptuneWarning,
            )

    @property
    def monitoring_namespace(self) -> str:
        return self._monitoring_namespace

    def _raise_if_stopped(self):
        if self._state == ContainerState.STOPPED:
            raise InactiveRunException(label=self._sys_id)

    def get_url(self) -> str:
        """Returns the URL that can be accessed within the browser"""
        return self._backend.get_run_url(
            run_id=self._id,
            workspace=self._workspace,
            project_name=self._project_name,
            sys_id=self._sys_id,
        )


def capture_only_if_non_interactive(mode) -> bool:
    if in_interactive() or in_notebook():
        if mode in {Mode.OFFLINE, Mode.SYNC, Mode.ASYNC}:
            warn_once(
                "The following monitoring options are disabled by default in interactive sessions:"
                " 'capture_stdout', 'capture_stderr', 'capture_traceback', and 'capture_hardware_metrics'."
                " To enable them, set each parameter to 'True' when initializing the run. The monitoring will"
                " continue until you call run.stop() or the kernel stops."
                " Also note: Your source files can only be tracked if you pass the path(s) to the 'source_code'"
                " argument. For help, see the Neptune docs: https://docs.neptune.ai/logging/source_code/",
                exception=NeptuneWarning,
            )
        return False
    return True


def generate_monitoring_namespace(*descriptors) -> str:
    return f"monitoring/{generate_hash(*descriptors, length=8)}"


def check_for_extra_kwargs(caller_name: str, kwargs: dict):
    legacy_kwargs = ("project_qualified_name", "backend")

    for name in legacy_kwargs:
        if name in kwargs:
            raise NeptunePossibleLegacyUsageException()

    if kwargs:
        first_key = next(iter(kwargs.keys()))
        raise TypeError(f"{caller_name}() got an unexpected keyword argument '{first_key}'")


def create_notebook_checkpoint(backend: NeptuneBackend) -> Tuple[Optional[str], Optional[str]]:
    notebook_id = os.getenv(NEPTUNE_NOTEBOOK_ID, None)
    notebook_path = os.getenv(NEPTUNE_NOTEBOOK_PATH, None)

    checkpoint_id = None
    if notebook_id is not None and notebook_path is not None:
        checkpoint_id = create_checkpoint(backend=backend, notebook_id=notebook_id, notebook_path=notebook_path)

    return notebook_id, checkpoint_id



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["StdoutCaptureBackgroundJob", "StderrCaptureBackgroundJob"]

from typing import (
    TYPE_CHECKING,
    Optional,
)

from neptune.internal.background_job import BackgroundJob
from neptune.internal.streams.std_stream_capture_logger import (
    StderrCaptureLogger,
    StdoutCaptureLogger,
)

if TYPE_CHECKING:
    from neptune.metadata_containers import MetadataContainer


class StdoutCaptureBackgroundJob(BackgroundJob):
    def __init__(self, attribute_name: str):
        self._attribute_name = attribute_name
        self._logger = None

    def start(self, container: "MetadataContainer"):
        self._logger = StdoutCaptureLogger(container, self._attribute_name)

    def stop(self):
        self._logger.close()

    def pause(self):
        self._logger.pause()

    def resume(self):
        self._logger.resume()

    def join(self, seconds: Optional[float] = None):
        pass


class StderrCaptureBackgroundJob(BackgroundJob):
    def __init__(self, attribute_name: str):
        self._attribute_name = attribute_name
        self._logger = None

    def start(self, container: "MetadataContainer"):
        self._logger = StderrCaptureLogger(container, self._attribute_name)

    def stop(self):
        self._logger.close()

    def pause(self):
        self._logger.pause()

    def resume(self):
        self._logger.resume()

    def join(self, seconds: Optional[float] = None):
        pass



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["ApiMethodWrapper", "SwaggerClientWrapper"]

from collections.abc import Callable
from typing import (
    Dict,
    Optional,
)

from bravado.client import SwaggerClient
from bravado.exception import HTTPError

from neptune.api.requests_utils import ensure_json_response
from neptune.common.exceptions import (
    NeptuneAuthTokenExpired,
    WritingToArchivedProjectException,
)
from neptune.exceptions import (
    NeptuneFieldCountLimitExceedException,
    NeptuneLimitExceedException,
)


class ApiMethodWrapper:
    def __init__(self, api_method):
        self._api_method = api_method

    @staticmethod
    def handle_neptune_http_errors(response, exception: Optional[HTTPError] = None):
        from neptune.management.exceptions import (
            ActiveProjectsLimitReachedException,
            IncorrectIdentifierException,
            ObjectNotFound,
            ProjectKeyCollision,
            ProjectKeyInvalid,
            ProjectNameCollision,
            ProjectNameInvalid,
            ProjectPrivacyRestrictedException,
            ProjectsLimitReached,
        )

        error_processors: Dict[str, Callable[[Dict], Exception]] = {
            "ATTRIBUTES_PER_EXPERIMENT_LIMIT_EXCEEDED": lambda response_body: NeptuneFieldCountLimitExceedException(
                limit=response_body.get("limit", "<unknown limit>"),
                container_type=response_body.get("experimentType", "object"),
                identifier=response_body.get("experimentQualifiedName", "<unknown identifier>"),
            ),
            "AUTHORIZATION_TOKEN_EXPIRED": lambda _: NeptuneAuthTokenExpired(),
            "EXPERIMENT_NOT_FOUND": lambda _: ObjectNotFound(),
            "INCORRECT_IDENTIFIER": lambda response_body: IncorrectIdentifierException(
                identifier=response_body.get("identifier", "<Unknown identifier>")
            ),
            "LIMIT_OF_PROJECTS_REACHED": lambda _: ProjectsLimitReached(),
            "PROJECT_KEY_COLLISION": lambda response_body: ProjectKeyCollision(
                key=response_body.get("key", "<unknown key>")
            ),
            "PROJECT_KEY_INVALID": lambda response_body: ProjectKeyInvalid(
                key=response_body.get("key", "<unknown key>"),
                reason=response_body.get("reason", "Unknown reason"),
            ),
            "PROJECT_NAME_COLLISION": lambda response_body: ProjectNameCollision(
                key=response_body.get("key", "<unknown key>")
            ),
            "PROJECT_NAME_INVALID": lambda response_body: ProjectNameInvalid(
                name=response_body.get("name", "<unknown name>")
            ),
            "VISIBILITY_RESTRICTED": lambda response_body: ProjectPrivacyRestrictedException(
                requested=response_body.get("requestedValue"),
                allowed=response_body.get("allowedValues"),
            ),
            "WORKSPACE_IN_READ_ONLY_MODE": lambda response_body: NeptuneLimitExceedException(
                reason=response_body.get("title", "Unknown reason")
            ),
            "PROJECT_LIMITS_EXCEEDED": lambda response_body: NeptuneLimitExceedException(
                reason=response_body.get("title", "Unknown reason")
            ),
            "LIMIT_OF_ACTIVE_PROJECTS_REACHED": lambda response_body: ActiveProjectsLimitReachedException(
                currentQuota=response_body.get("currentQuota", "<unknown quota>")
            ),
            "WRITE_ACCESS_DENIED_TO_ARCHIVED_PROJECT": lambda _: WritingToArchivedProjectException(),
        }

        body = ensure_json_response(response)
        error_type: Optional[str] = body.get("errorType")
        error_processor = error_processors.get(error_type)
        if error_processor:
            if exception:
                raise error_processor(body) from exception
            raise error_processor(body)

        if exception:
            raise exception

    def __call__(self, *args, **kwargs):
        try:
            future = self._api_method(*args, **kwargs)
            return FinishedApiResponseFuture(future.response())  # wait synchronously
        except HTTPError as e:
            self.handle_neptune_http_errors(e.response, exception=e)

    def __getattr__(self, item):
        return getattr(self._api_method, item)


class ApiWrapper:
    def __init__(self, api_obj):
        self._api_obj = api_obj

    def __getattr__(self, item):
        return ApiMethodWrapper(getattr(self._api_obj, item))


class FinishedApiResponseFuture:
    def __init__(self, response):
        self._response = response

    def response(self):
        return self._response


class SwaggerClientWrapper:
    def __init__(self, swagger_client: SwaggerClient):
        self._swagger_client = swagger_client
        self.api = ApiWrapper(swagger_client.api)
        self.swagger_spec = swagger_client.swagger_spec

    # For test purposes
    def __eq__(self, other):
        if isinstance(other, SwaggerClientWrapper):
            return self._swagger_client == other._swagger_client
        return False
